# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""
ODR (One Definition Rule) Helpers for CCCL Python Interop.

This module provides utilities to create wrapper functions for
device functions that are defined in Python and JIT compiled by Numba.

On the C++ side, these functions are declared as `extern "C"` functions with
void* parameters - the arguments types can not be known at C++ compile time.

Thus, the helpers in this module generate wrapper device functions that accept
void* arguments (matching C++ declarations), cast them to the correct
typed arguments, load/store values as needed, and call the original
function with properly typed arguments.

Example flow:
    User provides: def add(x: int32, y: int32) -> int32
    Wrapper signature: void(void*, void*, void*)  # x_ptr, y_ptr, result_ptr
    C++ sees: extern "C" void wrapped_add(void*, void*, void*);
"""

from __future__ import annotations

import enum
import itertools
import textwrap
import threading
from typing import TYPE_CHECKING

from numba import types
from numba.core.extending import intrinsic

from ._utils import sanitize_identifier

if TYPE_CHECKING:
    from numba.core.typing import Signature

# Global counter to generate unique symbol names even when the same function
# is used multiple times (e.g., as both selectors in `three_way_partition`).
_wrapper_name_counter = itertools.count()
_wrapper_name_lock = threading.Lock()

__all__ = [
    "create_op_void_ptr_wrapper",
    "create_stateful_op_void_ptr_wrapper",
    "create_advance_void_ptr_wrapper",
    "create_input_dereference_void_ptr_wrapper",
    "create_output_dereference_void_ptr_wrapper",
]


class _ArgMode(enum.Enum):
    """How a void* argument should be handled in wrapper codegen."""

    LOAD = "load"  # Cast to typed pointer, load value
    PTR = "ptr"  # Cast to typed pointer, pass pointer directly
    STORE = "store"  # Cast to typed pointer, store return value here
    # Unpack packed data pointers into array structs
    STATE = "state"


class _ArgSpec:
    """Specification for a wrapper argument."""

    __slots__ = ("numba_type", "mode")

    def __init__(self, numba_type, mode: _ArgMode):
        self.numba_type = numba_type
        self.mode = mode


def _build_numba_array_struct(context, builder, array_type, data_ptr, info):
    """Build a numba array struct from a data pointer and array info.

    Args:
        context: Numba codegen context
        builder: LLVM IR builder
        array_type: Numba Array type for the array
        data_ptr: LLVM value for the data pointer
        info: Dict with 'shape', 'itemsize', 'strides' for the array

    Returns:
        LLVM value representing the array struct
    """
    import llvmlite.ir as ir
    from numba.cuda.np.arrayobj import make_array, populate_array

    out_ary = make_array(array_type)(context, builder)

    populate_array(
        out_ary,
        data=data_ptr,
        shape=[ir.Constant(ir.IntType(64), info["shape"])],
        strides=[ir.Constant(ir.IntType(64), info["strides"])],
        itemsize=info["itemsize"],
        meminfo=None,
    )

    return out_ary._getvalue()


def _unpack_state_arrays(context, builder, packed_ptr, type_info_pairs):
    """Unpack packed data pointers into numba array structs.

    Args:
        context: Numba codegen context
        builder: LLVM IR builder
        packed_ptr: void* pointing to an array of data pointers
        type_info_pairs: List of (array_type, info) tuples

    Returns:
        List of LLVM values representing the unpacked array structs
    """
    import llvmlite.ir as ir

    # Cast void* to pointer-to-pointer (array of pointers)
    ptr_type = ir.IntType(64).as_pointer()
    base_ptr = builder.bitcast(packed_ptr, ptr_type.as_pointer())

    result = []
    for j, (array_type, info) in enumerate(type_info_pairs):
        # Load j-th pointer from the array and cast to correct type
        elem_ptr = builder.gep(base_ptr, [ir.Constant(ir.IntType(32), j)])
        dtype_llvm = context.get_value_type(array_type.dtype)
        typed_ptr_ptr = builder.bitcast(elem_ptr, dtype_llvm.as_pointer().as_pointer())
        data_ptr = builder.load(typed_ptr_ptr)

        # Build array struct from pointer
        array_val = _build_numba_array_struct(
            context, builder, array_type, data_ptr, info
        )
        result.append(array_val)

    return result


def _codegen_void_ptr_wrapper(
    context, builder, args, arg_specs, func_device, inner_sig
):
    """Generate LLVM IR for a void* wrapper function.

    This is the codegen implementation shared by all void* wrappers.
    It processes each argument according to its _ArgSpec mode, calls
    the inner function, and stores the result if needed.

    Args:
        context: Numba codegen context
        builder: LLVM IR builder
        args: LLVM values for the void* arguments
        arg_specs: List of _ArgSpec describing each argument
        func_device: The device function to call
        inner_sig: Numba signature for the inner function

    Returns:
        LLVM dummy value (for void return)
    """

    input_vals = []
    state_array_vals = []
    ret_ptr = None

    for i, (arg, spec) in enumerate(zip(args, arg_specs)):
        match spec.mode:
            case _ArgMode.LOAD:
                # Cast void* to typed pointer and load value
                llvm_type = context.get_value_type(spec.numba_type)
                typed_ptr = builder.bitcast(arg, llvm_type.as_pointer())
                val = builder.load(typed_ptr)
                input_vals.append(val)
            case _ArgMode.PTR:
                # Cast void* to typed pointer, pass pointer directly
                llvm_type = context.get_value_type(spec.numba_type.dtype)
                typed_ptr = builder.bitcast(arg, llvm_type.as_pointer())
                input_vals.append(typed_ptr)
            case _ArgMode.STORE:
                # Cast void* to typed pointer for storing result
                llvm_type = context.get_value_type(spec.numba_type)
                ret_ptr = builder.bitcast(arg, llvm_type.as_pointer())
            case _ArgMode.STATE:
                # Cast void* to a packed array of pointers and unpack them
                array_vals = _unpack_state_arrays(
                    context, builder, arg, spec.numba_type
                )
                state_array_vals.extend(array_vals)
            case _:
                raise ValueError(f"Invalid arg mode: {spec.mode}")

    # Prepend state arrays at the beginning (inner_sig expects state args first)
    input_vals = state_array_vals + input_vals

    # Call the inner function
    cres = context.compile_subroutine(builder, func_device, inner_sig, caching=False)
    result = context.call_internal(builder, cres.fndesc, inner_sig, input_vals)

    # Store result if needed
    if ret_ptr is not None:
        builder.store(result, ret_ptr)

    return context.get_dummy_value()


def _create_void_ptr_wrapper(
    func, name: str, arg_specs: list[_ArgSpec], inner_sig: "Signature"
):
    """
    Given a function and a list of _ArgSpec, create a wrapper function
    that takes all void* arguments, bitcasts them to the
    appropriate typed pointers, and calls the inner function with
    the typed arguments. Each void* argument is handled according
    to its _ArgSpec.

    Args:
        func: The function to wrap (will be compiled as device function)
        name: Base name for the wrapper function
        arg_specs: List of _ArgSpec describing each void* argument
        inner_sig: Numba signature for the inner function call

    Returns:
        Tuple of (wrapper_func, wrapper_sig)
    """
    from numba.cuda import jit as cuda_jit

    # Wrap function as device function
    func_device = cuda_jit(device=True)(func)

    # Generate argument names and signature
    arg_names = [f"arg_{i}" for i in range(len(arg_specs))]
    arg_str = ", ".join(arg_names)
    void_sig = types.void(*(types.voidptr for _ in arg_specs))

    # Create unique wrapper name using global counter
    sanitized_name = sanitize_identifier(name)
    if not sanitized_name.isidentifier():
        raise ValueError(
            f"Function name '{name}' cannot be sanitized into a valid identifier"
        )

    for arg_name in arg_names:
        if not arg_name.isidentifier():
            raise ValueError(
                f"Invalid argument name '{arg_name}' - must be a valid identifier"
            )
    with _wrapper_name_lock:
        unique_suffix = next(_wrapper_name_counter)
    wrapper_name = f"wrapped_{sanitized_name}_{unique_suffix}"

    # We need exec() here because Numba's @intrinsic decorator requires:
    # 1. A function with a specific signature visible at parse time
    # 2. The number of arguments must match the wrapper signature
    # The actual codegen logic is in _codegen_void_ptr_wrapper - this just
    # creates the minimal intrinsic shell that delegates to it.
    wrapper_src = textwrap.dedent(f"""
    @intrinsic
    def impl(typingctx, {arg_str}):
        def codegen(context, builder, impl_sig, args):
            return codegen_helper(context, builder, args, arg_specs, func_device, inner_sig)
        return void_sig, codegen

    def {wrapper_name}({arg_str}):
        return impl({arg_str})
    """)

    local_dict = {
        "intrinsic": intrinsic,
        "void_sig": void_sig,
        "arg_specs": arg_specs,
        "func_device": func_device,
        "inner_sig": inner_sig,
        "codegen_helper": _codegen_void_ptr_wrapper,
    }
    exec(wrapper_src, {}, local_dict)

    wrapper_func = local_dict[wrapper_name]
    wrapper_func.__globals__.update(local_dict)

    return wrapper_func, void_sig


def create_op_void_ptr_wrapper(op, sig: "Signature"):
    """Creates a wrapper function for user-defined operators like unary or binary operators.

    The wrapper takes N+1 arguments where N is the number of input arguments to `op`, the last
    argument is a pointer to the result.
    """
    arg_specs = [_ArgSpec(t, _ArgMode.LOAD) for t in sig.args]
    arg_specs.append(_ArgSpec(sig.return_type, _ArgMode.STORE))
    return _create_void_ptr_wrapper(op, op.__name__, arg_specs, sig)


def create_stateful_op_void_ptr_wrapper(
    op, sig: "Signature", state_array_types, state_info
):
    """Creates a wrapper function for a stateful operator with void* arguments.

    The wrapper takes N+2 void* arguments:
    - states_ptr: pointer to packed array of data pointers for state arrays
    - N input args: one for each regular input argument
    - result: pointer where result is stored

    Args:
        op: The user's callable operator
        sig: The signature of the operator (state_array1, state_array2, ..., regular_arg1, regular_arg2, ...) -> return_type
        state_array_types: List/tuple of numba Array types for the state parameters
        state_info: List/tuple of dicts with 'shape', 'itemsize', 'strides' for each state array

    Returns:
        Tuple of (wrapper_func, wrapper_sig)
    """
    num_states = len(state_array_types)

    # Build arg_specs: states_ptr + regular inputs + result
    # The packed state arrays spec goes first, then regular LOAD args, then STORE for result
    # numba_type is a list of (array_type, info) tuples
    type_info_pairs = list(zip(state_array_types, state_info))
    arg_specs = [_ArgSpec(type_info_pairs, _ArgMode.STATE)]
    for i in range(num_states, len(sig.args)):
        arg_specs.append(_ArgSpec(sig.args[i], _ArgMode.LOAD))
    arg_specs.append(_ArgSpec(sig.return_type, _ArgMode.STORE))

    return _create_void_ptr_wrapper(op, op.__name__, arg_specs, sig)


def create_advance_void_ptr_wrapper(advance_fn, state_ptr_type):
    """Creates a wrapper function for iterator advance method.

    The wrapper takes 2 void* arguments:
    - state pointer
    - offset pointer (points to uint64 value)
    """
    arg_specs = [
        _ArgSpec(state_ptr_type, _ArgMode.PTR),
        _ArgSpec(types.uint64, _ArgMode.LOAD),  # uint64 is the offset type
    ]
    inner_sig = types.void(state_ptr_type, types.uint64)
    return _create_void_ptr_wrapper(
        advance_fn, advance_fn.__name__, arg_specs, inner_sig
    )


def create_input_dereference_void_ptr_wrapper(deref_fn, state_ptr_type, value_type):
    """Creates a wrapper function for input iterator dereference method.

    The wrapper takes 2 void* arguments:
    - state pointer
    - result pointer (function writes result here)
    """
    arg_specs = [
        _ArgSpec(state_ptr_type, _ArgMode.PTR),
        _ArgSpec(types.CPointer(value_type), _ArgMode.PTR),
    ]
    inner_sig = types.void(state_ptr_type, types.CPointer(value_type))
    return _create_void_ptr_wrapper(deref_fn, deref_fn.__name__, arg_specs, inner_sig)


def create_output_dereference_void_ptr_wrapper(deref_fn, state_ptr_type, value_type):
    """Creates a wrapper function for output iterator dereference method.

    The wrapper takes 2 void* arguments:
    - state pointer
    - value pointer (value to write)
    """
    arg_specs = [
        _ArgSpec(state_ptr_type, _ArgMode.PTR),
        _ArgSpec(value_type, _ArgMode.LOAD),
    ]
    inner_sig = types.void(state_ptr_type, value_type)
    return _create_void_ptr_wrapper(deref_fn, deref_fn.__name__, arg_specs, inner_sig)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from typing import Callable

from .. import _bindings, types
from .. import _cccl_interop as cccl
from .._caching import cache_with_registered_key_functions
from .._cccl_interop import call_build, set_cccl_iterator_state
from .._utils import protocols
from .._utils.temp_storage_buffer import TempStorageBuffer
from ..op import OpAdapter, make_op_adapter
from ..typing import DeviceArrayLike, IteratorT, Operator


class _ThreeWayPartition:
    __slots__ = [
        "build_result",
        "d_in_cccl",
        "d_first_part_out_cccl",
        "d_second_part_out_cccl",
        "d_unselected_out_cccl",
        "d_num_selected_out_cccl",
        "select_first_part_op_cccl",
        "select_second_part_op_cccl",
    ]

    def __init__(
        self,
        d_in: DeviceArrayLike | IteratorT,
        d_first_part_out: DeviceArrayLike | IteratorT,
        d_second_part_out: DeviceArrayLike | IteratorT,
        d_unselected_out: DeviceArrayLike | IteratorT,
        d_num_selected_out: DeviceArrayLike | IteratorT,
        select_first_part_op: OpAdapter,
        select_second_part_op: OpAdapter,
    ):
        self.d_in_cccl = cccl.to_cccl_input_iter(d_in)
        self.d_first_part_out_cccl = cccl.to_cccl_output_iter(d_first_part_out)
        self.d_second_part_out_cccl = cccl.to_cccl_output_iter(d_second_part_out)
        self.d_unselected_out_cccl = cccl.to_cccl_output_iter(d_unselected_out)
        self.d_num_selected_out_cccl = cccl.to_cccl_output_iter(d_num_selected_out)

        # Compile ops - partition predicates return uint8 (boolean)
        value_type = cccl.get_value_type(d_in)
        self.select_first_part_op_cccl = select_first_part_op.compile(
            (value_type,), types.uint8
        )
        self.select_second_part_op_cccl = select_second_part_op.compile(
            (value_type,), types.uint8
        )

        self.build_result = call_build(
            _bindings.DeviceThreeWayPartitionBuildResult,
            self.d_in_cccl,
            self.d_first_part_out_cccl,
            self.d_second_part_out_cccl,
            self.d_unselected_out_cccl,
            self.d_num_selected_out_cccl,
            self.select_first_part_op_cccl,
            self.select_second_part_op_cccl,
        )

    def __call__(
        self,
        temp_storage,
        d_in,
        d_first_part_out,
        d_second_part_out,
        d_unselected_out,
        d_num_selected_out,
        select_first_part_op: Callable | OpAdapter,
        select_second_part_op: Callable | OpAdapter,
        num_items: int,
        stream=None,
    ):
        set_cccl_iterator_state(self.d_in_cccl, d_in)
        set_cccl_iterator_state(self.d_first_part_out_cccl, d_first_part_out)
        set_cccl_iterator_state(self.d_second_part_out_cccl, d_second_part_out)
        set_cccl_iterator_state(self.d_unselected_out_cccl, d_unselected_out)
        set_cccl_iterator_state(self.d_num_selected_out_cccl, d_num_selected_out)

        first_op_adapter = make_op_adapter(select_first_part_op)
        second_op_adapter = make_op_adapter(select_second_part_op)
        self.select_first_part_op_cccl.state = first_op_adapter.get_state()
        self.select_second_part_op_cccl.state = second_op_adapter.get_state()

        stream_handle = protocols.validate_and_get_stream(stream)

        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            d_temp_storage = protocols.get_data_pointer(temp_storage)

        temp_storage_bytes = self.build_result.compute(
            d_temp_storage,
            temp_storage_bytes,
            self.d_in_cccl,
            self.d_first_part_out_cccl,
            self.d_second_part_out_cccl,
            self.d_unselected_out_cccl,
            self.d_num_selected_out_cccl,
            self.select_first_part_op_cccl,
            self.select_second_part_op_cccl,
            num_items,
            stream_handle,
        )
        return temp_storage_bytes


@cache_with_registered_key_functions
def make_three_way_partition(
    d_in: DeviceArrayLike | IteratorT,
    d_first_part_out: DeviceArrayLike | IteratorT,
    d_second_part_out: DeviceArrayLike | IteratorT,
    d_unselected_out: DeviceArrayLike | IteratorT,
    d_num_selected_out: DeviceArrayLike | IteratorT,
    select_first_part_op: Operator,
    select_second_part_op: Operator,
):
    """
    Computes a device-wide three-way partition using the specified unary ``select_first_part_op`` and ``select_second_part_op`` operators.

    Example:
        Below, ``make_three_way_partition`` is used to create a three-way partition object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/partition/three_way_partition_object.py
            :language: python
            :start-after: # example-begin

    Args:
        d_in: Device array or iterator containing the input sequence of data items
        d_first_part_out: Device array or iterator to store the first part of the output
        d_second_part_out: Device array or iterator to store the second part of the output
        d_unselected_out: Device array or iterator to store the unselected items
        d_num_selected_out: Device array to store the number of items selected. The total number of items selected by ``select_first_part_op`` and ``select_second_part_op`` is stored in ``d_num_selected_out[0]`` and ``d_num_selected_out[1]``, respectively.
        select_first_part_op: Unary operator to select the first part.
            The signature is ``(T) -> uint8``, where ``T`` is the input data type.
            Returns 1 (selected) or 0 (not selected).
            Can reference device arrays as globals/closures - they will be automatically captured.
        select_second_part_op: Unary operator to select the second part.
            The signature is ``(T) -> uint8``, where ``T`` is the input data type.
            Returns 1 (selected) or 0 (not selected).
            Can reference device arrays as globals/closures - they will be automatically captured.

    Returns:
        A callable object that can be used to perform the three-way partition
    """
    first_op_adapter = make_op_adapter(select_first_part_op)
    second_op_adapter = make_op_adapter(select_second_part_op)

    return _ThreeWayPartition(
        d_in,
        d_first_part_out,
        d_second_part_out,
        d_unselected_out,
        d_num_selected_out,
        first_op_adapter,
        second_op_adapter,
    )


def three_way_partition(
    d_in: DeviceArrayLike | IteratorT,
    d_first_part_out: DeviceArrayLike | IteratorT,
    d_second_part_out: DeviceArrayLike | IteratorT,
    d_unselected_out: DeviceArrayLike | IteratorT,
    d_num_selected_out: DeviceArrayLike | IteratorT,
    select_first_part_op: Operator,
    select_second_part_op: Operator,
    num_items: int,
    stream=None,
):
    """
    Performs device-wide three-way partition. Given an input sequence of data items, it partitions the items into three parts:
    - The first part is selected by the ``select_first_part_op`` operator.
    - The second part is selected by the ``select_second_part_op`` operator.
    - The unselected items are not selected by either operator.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``three_way_partition`` is used to partition a sequence of integers into three parts.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/partition/three_way_partition_basic.py
            :language: python
            :start-after: # example-begin

    Args:
        d_in: Device array or iterator containing the input sequence of data items
        d_first_part_out: Device array or iterator to store the first part of the output
        d_second_part_out: Device array or iterator to store the second part of the output
        d_unselected_out: Device array or iterator to store the unselected items
        d_num_selected_out: Device array to store the number of items selected. The total number of items selected by ``select_first_part_op`` and ``select_second_part_op`` is stored in ``d_num_selected_out[0]`` and ``d_num_selected_out[1]``, respectively.
        select_first_part_op: Unary operator to select the first part.
            The signature is ``(T) -> uint8``, where ``T`` is the input data type.
            Returns 1 (selected) or 0 (not selected).
        select_second_part_op: Unary operator to select the second part.
            The signature is ``(T) -> uint8``, where ``T`` is the input data type.
            Returns 1 (selected) or 0 (not selected).
        num_items: Number of items to partition
        stream: CUDA stream for the operation (optional)
    """
    # Create adapters to support stateful ops
    first_op_adapter = make_op_adapter(select_first_part_op)
    second_op_adapter = make_op_adapter(select_second_part_op)

    partitioner = make_three_way_partition(
        d_in,
        d_first_part_out,
        d_second_part_out,
        d_unselected_out,
        d_num_selected_out,
        first_op_adapter,
        second_op_adapter,
    )
    tmp_storage_bytes = partitioner(
        None,
        d_in,
        d_first_part_out,
        d_second_part_out,
        d_unselected_out,
        d_num_selected_out,
        first_op_adapter,
        second_op_adapter,
        num_items,
        stream,
    )
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    partitioner(
        tmp_storage,
        d_in,
        d_first_part_out,
        d_second_part_out,
        d_unselected_out,
        d_num_selected_out,
        first_op_adapter,
        second_op_adapter,
        num_items,
        stream,
    )



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import hashlib
import json
import os

_ENABLE_CACHE = "CCCL_ENABLE_CACHE" in os.environ
if _ENABLE_CACHE:
    _CACHE_LOCATION = os.path.join(os.path.expanduser("~"), ".cache", "cccl")
    if not os.path.exists(_CACHE_LOCATION):
        os.makedirs(_CACHE_LOCATION)


# We use
# json.dumps to serialize args/kwargs to a string
# hashlib to compute the hash
def json_hash(*args, **kwargs):
    hasher = hashlib.sha1()
    hasher.update(json.dumps([args, kwargs]).encode("utf-8"))
    return hasher.hexdigest()


def disk_cache(func):
    def cacher(*args, **kwargs):
        if _ENABLE_CACHE:
            # compute hash(args, kwargs)
            h = json_hash(*args, **kwargs)
            # if file exist...
            if os.path.isfile(os.path.join(_CACHE_LOCATION, h)):
                # open it
                with open(os.path.join(_CACHE_LOCATION, h), "rb") as f:
                    out = json.load(f)
                # return cache
                return out
            else:
                # compute output
                out = func(*args, **kwargs)
                # store to file
                with open(os.path.join(_CACHE_LOCATION, h), "wb") as f:
                    json.dump(out, f)
                return out
        else:
            return func(*args, **kwargs)

    return cacher



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from . import block, warp
from ._types import StatefulFunction

__all__ = ["block", "warp", "StatefulFunction"]



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from .include_paths import get_include_paths

__all__ = ["__version__", "get_include_paths"]



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""PermutationIterator implementation."""

from __future__ import annotations

from textwrap import dedent

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir, make_variable_declaration
from ._base import IteratorBase, compose_iterator_states
from ._common import CUDA_PREAMBLE, ensure_iterator


class PermutationIterator(IteratorBase):
    """
    Iterator that accesses values through an index mapping.

    At position i, yields values[indices[i]].

    Similar to `thrust::permutation_iterator <https://nvidia.github.io/cccl/thrust/api/classthrust_1_1permutation__iterator.html>`_.

    Example:
        The code snippet below demonstrates accessing values through an index mapping.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/permutation_iterator_basic.py
            :language: python
            :start-after: # example-begin
    """

    __slots__ = [
        "_values",
        "_indices",
        "_values_offset",
        "_indices_offset",
    ]

    def __init__(
        self,
        values,
        indices,
    ):
        """
        Create a permutation iterator.

        Args:
            values: Iterator or array providing the values to be permuted
            indices: Iterator or array providing the indices for permutation
        """
        # Wrap arrays in PointerIterator
        self._values = ensure_iterator(values)
        self._indices = ensure_iterator(indices)

        # Compose states from both iterators
        state_bytes, state_alignment, offsets = compose_iterator_states(
            [self._values, self._indices]
        )
        self._values_offset = offsets[0]
        self._indices_offset = offsets[1]

        super().__init__(
            state_bytes=state_bytes,
            state_alignment=state_alignment,
            value_type=self._values.value_type,
        )

    def _make_advance_op(self) -> Op:
        """Provide Op for advance that only advances indices iterator."""
        child_op = self._indices.get_advance_op()
        symbol = self._make_advance_symbol()

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {child_op.name}(void* state, void* offset);

            extern "C" __device__ void {symbol}(void* state, void* offset) {{
                char* indices_state = static_cast<char*>(state) + {self._indices_offset};
                {child_op.name}(indices_state, offset);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[child_op.ltoir, *child_op.extra_ltoirs],
        )

    def _make_input_deref_op(self) -> Op | None:
        """Provide Op for input deref that reads index then accesses values."""
        indices_deref_op = self._indices.get_input_deref_op()
        if indices_deref_op is None:
            raise ValueError("Indices iterator must support input dereference")

        values_deref_op = self._values.get_input_deref_op()
        if values_deref_op is None:
            return None

        # Also need values advance for random access
        values_advance_op = self._values.get_advance_op()

        symbol = self._make_input_deref_symbol()
        idx_decl = make_variable_declaration(self._indices.value_type, "idx")
        values_state_size = len(bytes(memoryview(self._values.state)))

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {indices_deref_op.name}(void* state, void* result);
            extern "C" __device__ void {values_advance_op.name}(void* state, void* offset);
            extern "C" __device__ void {values_deref_op.name}(void* state, void* result);

            extern "C" __device__ void {symbol}(void* state, void* result) {{
                char* values_state = static_cast<char*>(state) + {self._values_offset};
                char* indices_state = static_cast<char*>(state) + {self._indices_offset};

                {idx_decl}
                {indices_deref_op.name}(indices_state, &idx);

                alignas({self._values.state_alignment}) char temp_values[{values_state_size}];
                memcpy(temp_values, values_state, {values_state_size});

                uint64_t offset = static_cast<uint64_t>(idx);
                {values_advance_op.name}(temp_values, &offset);
                {values_deref_op.name}(temp_values, result);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[
                values_advance_op.ltoir,
                *values_advance_op.extra_ltoirs,
                indices_deref_op.ltoir,
                *indices_deref_op.extra_ltoirs,
                values_deref_op.ltoir,
                *values_deref_op.extra_ltoirs,
            ],
        )

    def _make_output_deref_op(self) -> Op | None:
        """Provide Op for output deref that reads index then writes to values."""
        indices_deref_op = self._indices.get_input_deref_op()
        if indices_deref_op is None:
            raise ValueError("Indices iterator must support input dereference")

        values_deref_op = self._values.get_output_deref_op()
        if values_deref_op is None:
            return None

        # Also need values advance for random access
        values_advance_op = self._values.get_advance_op()

        symbol = self._make_output_deref_symbol()
        idx_decl = make_variable_declaration(self._indices.value_type, "idx")
        values_state_size = len(bytes(memoryview(self._values.state)))

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {indices_deref_op.name}(void* state, void* result);
            extern "C" __device__ void {values_advance_op.name}(void* state, void* offset);
            extern "C" __device__ void {values_deref_op.name}(void* state, void* value);

            extern "C" __device__ void {symbol}(void* state, void* value) {{
                char* values_state = static_cast<char*>(state) + {self._values_offset};
                char* indices_state = static_cast<char*>(state) + {self._indices_offset};

                {idx_decl}
                {indices_deref_op.name}(indices_state, &idx);

                alignas({self._values.state_alignment}) char temp_values[{values_state_size}];
                memcpy(temp_values, values_state, {values_state_size});

                uint64_t offset = static_cast<uint64_t>(idx);
                {values_advance_op.name}(temp_values, &offset);
                {values_deref_op.name}(temp_values, value);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[
                values_advance_op.ltoir,
                *values_advance_op.extra_ltoirs,
                indices_deref_op.ltoir,
                *indices_deref_op.extra_ltoirs,
                values_deref_op.ltoir,
                *values_deref_op.extra_ltoirs,
            ],
        )

    @property
    def children(self):
        return (self._values, self._indices)

    def __add__(self, offset: int) -> "PermutationIterator":
        """Advance the indices iterator by offset, keeping values at base."""
        return PermutationIterator(
            self._values,  # values stays at base for random access
            self._indices + offset,  # only indices advances  # type: ignore[operator]
        )

    @property
    def kind(self):
        """Return a hashable kind for caching purposes."""
        return ("PermutationIterator", self._values.kind, self._indices.kind)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from functools import cache

from .._caching import cache_with_registered_key_functions
from .._cpp_compile import compile_cpp_to_ltoir
from .._utils.temp_storage_buffer import TempStorageBuffer
from ..iterators import DiscardIterator
from ..op import OpAdapter, RawOp, make_op_adapter
from ..typing import DeviceArrayLike, IteratorT, Operator
from ._three_way_partition import make_three_way_partition


@cache
def _always_false_op():
    source = """
extern "C" __device__ void always_false(void*, void* result) {{
    *static_cast<bool*>(result) = false;
}}
"""
    ltoir = compile_cpp_to_ltoir(source)
    return RawOp(ltoir=ltoir, name="always_false")


class _Select:
    __slots__ = ["partitioner", "discard_second", "discard_unselected", "false_op"]

    def __init__(
        self,
        d_in: DeviceArrayLike | IteratorT,
        d_out: DeviceArrayLike | IteratorT,
        d_num_selected_out: DeviceArrayLike,
        cond: OpAdapter,
    ):
        # Create discard iterators for unused outputs, using d_out as reference
        # to match the input/output type
        self.discard_second = DiscardIterator(d_out)
        self.discard_unselected = DiscardIterator(d_out)

        # Create adapter for the always-false second predicate
        self.false_op = _always_false_op()

        # Use three_way_partition internally
        self.partitioner = make_three_way_partition(
            d_in,
            d_out,  # first_part_out - this is where selected items go
            self.discard_second,  # second_part_out - discarded
            self.discard_unselected,  # unselected_out - discarded
            d_num_selected_out,
            cond,  # select_first_part_op - user's select condition
            self.false_op,  # select_second_part_op - always false
        )

    def __call__(
        self,
        temp_storage,
        d_in,
        d_out,
        d_num_selected_out,
        cond,
        num_items: int,
        stream=None,
    ):
        return self.partitioner(
            temp_storage,
            d_in,
            d_out,
            self.discard_second,
            self.discard_unselected,
            d_num_selected_out,
            make_op_adapter(cond),
            self.false_op,
            num_items,
            stream,
        )


@cache_with_registered_key_functions
def make_select(
    d_in: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    d_num_selected_out: DeviceArrayLike,
    cond: Operator,
):
    """
    Create a select object that can be called to select elements matching a condition.

    This is the object-oriented API that allows explicit control over temporary
    storage allocation. For simpler usage, consider using :func:`select`.

    Example:
        Below, ``make_select`` is used to create a select object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/select/select_object.py
            :language: python
            :start-after: # example-begin

    Args:
        d_in: Device array or iterator containing the input sequence of data items.
        d_out: Device array or iterator to store the selected output items.
        d_num_selected_out: Device array to store the number of items that passed the selection.
            The count is stored in ``d_num_selected_out[0]``.
        cond: Selection condition (predicate).
            The signature is ``(T) -> uint8``, where ``T`` is the input data type.
            Returns 1 (selected) or 0 (not selected).

    Returns:
        A callable object that performs the selection operation.
    """
    cond_adapter = make_op_adapter(cond)
    # Note: _Select internally calls make_three_way_partition which will
    # normalize the cond. But we've already normalized it, so the Op
    # will be passed through make_op unchanged.
    return _Select(d_in, d_out, d_num_selected_out, cond_adapter)


def select(
    d_in: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    d_num_selected_out: DeviceArrayLike,
    cond: Operator,
    num_items: int,
    stream=None,
):
    """
    Performs device-wide selection of elements based on a condition.

    Given an input sequence, this function selects all elements for which the condition
    function ``cond`` returns true (non-zero) and writes them to the output in a
    compacted form. The number of selected elements is written to ``d_num_selected_out[0]``.

    This function automatically handles temporary storage allocation and execution.

    The ``cond`` function can reference device arrays as globals or closures - they will
    be automatically captured as state arrays, enabling stateful operations like counting.

    Example:
        Below, ``select`` is used to select even numbers from an input array:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/select/select_basic.py
            :language: python
            :start-after: # example-begin

        You can also use iterators for more complex selection patterns:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/select/select_with_iterator.py
            :language: python
            :start-after: # example-begin

    Args:
        d_in: Device array or iterator containing the input sequence of data items.
        d_out: Device array or iterator to store the selected output items.
        d_num_selected_out: Device array to store the number of items that passed the selection.
            The count is stored in ``d_num_selected_out[0]``.
        cond: Selection condition (predicate).
            The signature is ``(T) -> uint8``, where ``T`` is the input data type.
            Returns 1 (selected) or 0 (not selected).
            Can reference device arrays as globals/closures - they will be automatically captured.
        num_items: Number of items in the input sequence.
        stream: CUDA stream to use for the operation (optional).
    """
    # Create adapter to support stateful ops
    cond_adapter = make_op_adapter(cond)
    selector = make_select(d_in, d_out, d_num_selected_out, cond_adapter)

    tmp_storage_bytes = selector(
        None,
        d_in,
        d_out,
        d_num_selected_out,
        cond_adapter,
        num_items,
        stream,
    )
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    selector(
        tmp_storage,
        d_in,
        d_out,
        d_num_selected_out,
        cond_adapter,
        num_items,
        stream,
    )



# Copyright (c) 2024-2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from enum import Enum
from typing import Tuple

from ..._caching import cache_with_registered_key_functions
from ..._utils.protocols import get_dtype
from ...typing import DeviceArrayLike


class SortOrder(Enum):
    ASCENDING = 0
    DESCENDING = 1


class DoubleBuffer:
    def __init__(self, d_current: DeviceArrayLike, d_alternate: DeviceArrayLike):
        self.d_buffers = [d_current, d_alternate]
        self.selector = 0

    def current(self):
        return self.d_buffers[self.selector]

    def alternate(self):
        return self.d_buffers[1 - self.selector]


def _get_arrays(
    d_in_keys: DeviceArrayLike | DoubleBuffer,
    d_out_keys: DeviceArrayLike | None,
    d_in_values: DeviceArrayLike | DoubleBuffer | None,
    d_out_values: DeviceArrayLike | None,
) -> Tuple[DeviceArrayLike, DeviceArrayLike, DeviceArrayLike, DeviceArrayLike]:
    if isinstance(d_in_keys, DoubleBuffer):
        d_in_keys_array = d_in_keys.current()
        d_out_keys_array = d_in_keys.alternate()

        if d_in_values is not None:
            assert isinstance(d_in_values, DoubleBuffer)
            d_in_values_array = d_in_values.current()
            d_out_values_array = d_in_values.alternate()
        else:
            d_in_values_array = None
            d_out_values_array = None
    else:
        d_in_keys_array = d_in_keys
        d_in_values_array = d_in_values
        d_out_keys_array = d_out_keys
        d_out_values_array = d_out_values

    return d_in_keys_array, d_out_keys_array, d_in_values_array, d_out_values_array


# DoubleBuffer: extract dtype from current buffer
cache_with_registered_key_functions.register(
    DoubleBuffer, lambda buf: get_dtype(buf.current())
)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from typing import Callable

from .. import _bindings
from .. import _cccl_interop as cccl
from .._caching import cache_with_registered_key_functions
from .._cccl_interop import set_cccl_iterator_state
from .._utils import protocols
from ..op import OpAdapter, make_op_adapter
from ..typing import DeviceArrayLike, IteratorT, Operator


class _UnaryTransform:
    __slots__ = ["d_in_cccl", "d_out_cccl", "op_cccl", "build_result"]

    def __init__(
        self,
        d_in: DeviceArrayLike | IteratorT,
        d_out: DeviceArrayLike | IteratorT,
        op: OpAdapter,
    ):
        self.d_in_cccl = cccl.to_cccl_input_iter(d_in)
        self.d_out_cccl = cccl.to_cccl_output_iter(d_out)

        # Compile the op with input/output types
        in_type = cccl.get_value_type(d_in)
        out_type = cccl.get_value_type(d_out)
        self.op_cccl = op.compile((in_type,), out_type)

        self.build_result = cccl.call_build(
            _bindings.DeviceUnaryTransform,
            self.d_in_cccl,
            self.d_out_cccl,
            self.op_cccl,
        )

    def __call__(
        self,
        d_in,
        d_out,
        op: Callable | OpAdapter,
        num_items: int,
        stream=None,
    ):
        op_adapter = make_op_adapter(op)

        set_cccl_iterator_state(self.d_in_cccl, d_in)
        set_cccl_iterator_state(self.d_out_cccl, d_out)
        self.op_cccl.state = op_adapter.get_state()

        stream_handle = protocols.validate_and_get_stream(stream)
        self.build_result.compute(
            self.d_in_cccl,
            self.d_out_cccl,
            num_items,
            self.op_cccl,
            stream_handle,
        )
        return None


class _BinaryTransform:
    __slots__ = [
        "d_in1_cccl",
        "d_in2_cccl",
        "d_out_cccl",
        "op_cccl",
        "build_result",
    ]

    def __init__(
        self,
        d_in1: DeviceArrayLike | IteratorT,
        d_in2: DeviceArrayLike | IteratorT,
        d_out: DeviceArrayLike | IteratorT,
        op: OpAdapter,
    ):
        self.d_in1_cccl = cccl.to_cccl_input_iter(d_in1)
        self.d_in2_cccl = cccl.to_cccl_input_iter(d_in2)
        self.d_out_cccl = cccl.to_cccl_output_iter(d_out)

        # Compile the op with input/output types
        in1_type = cccl.get_value_type(d_in1)
        in2_type = cccl.get_value_type(d_in2)
        out_type = cccl.get_value_type(d_out)
        self.op_cccl = op.compile((in1_type, in2_type), out_type)

        self.build_result = cccl.call_build(
            _bindings.DeviceBinaryTransform,
            self.d_in1_cccl,
            self.d_in2_cccl,
            self.d_out_cccl,
            self.op_cccl,
        )

    def __call__(
        self,
        d_in1,
        d_in2,
        d_out,
        op: Callable | OpAdapter,
        num_items: int,
        stream=None,
    ):
        set_cccl_iterator_state(self.d_in1_cccl, d_in1)
        set_cccl_iterator_state(self.d_in2_cccl, d_in2)
        set_cccl_iterator_state(self.d_out_cccl, d_out)

        op_adapter = make_op_adapter(op)
        self.op_cccl.state = op_adapter.get_state()

        stream_handle = protocols.validate_and_get_stream(stream)
        self.build_result.compute(
            self.d_in1_cccl,
            self.d_in2_cccl,
            self.d_out_cccl,
            num_items,
            self.op_cccl,
            stream_handle,
        )
        return None


@cache_with_registered_key_functions
def make_unary_transform(
    d_in: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    op: Operator,
):
    """
    Create a unary transform object that can be called to apply a transformation
    to each element of the input according to the unary operation ``op``.

    This is the object-oriented API that allows explicit control over temporary
    storage allocation. For simpler usage, consider using :func:`unary_transform`.

    Example:
        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/transform/unary_transform_object.py
           :language: python
           :start-after: # example-begin


    Args:
        d_in: Device array or iterator containing the input sequence of data items.
        d_out: Device array or iterator to store the result of the transformation.
        op: Unary operation to apply to each element.
            The signature is ``(T) -> U``, where ``T`` is
            the input data type and ``U`` is the output data type.

    Returns:
        A callable object that performs the transformation.
    """
    op_adapter = make_op_adapter(op)
    return _UnaryTransform(d_in, d_out, op_adapter)


@cache_with_registered_key_functions
def make_binary_transform(
    d_in1: DeviceArrayLike | IteratorT,
    d_in2: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    op: Operator,
):
    """
    Create a binary transform object that can be called to apply a transformation
    to the given pair of input sequences according to the binary operation ``op``.

    This is the object-oriented API that allows explicit control over temporary
    storage allocation. For simpler usage, consider using :func:`binary_transform`.

    Example:
        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/transform/binary_transform_object.py
           :language: python
           :start-after: # example-begin


    Args:
        d_in1: Device array or iterator containing the first input sequence of data items.
        d_in2: Device array or iterator containing the second input sequence of data items.
        d_out: Device array or iterator to store the result of the transformation.
        op: Binary operation.
            The signature is ``(T1, T2) -> U``, where ``T1`` and ``T2`` are the input data types and
            ``U`` is the output data type.

    Returns:
        A callable object that performs the transformation.
    """
    op_adapter = make_op_adapter(op)
    return _BinaryTransform(d_in1, d_in2, d_out, op_adapter)


def unary_transform(
    d_in: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    op: Operator,
    num_items: int,
    stream=None,
):
    """
    Performs device-wide unary transform.

    This function automatically handles temporary storage allocation and execution.

    The ``op`` function can reference device arrays as globals or closures - they will
    be automatically captured as state arrays, enabling stateful operations like counting.

    Example:
        Below, ``unary_transform`` is used to apply a transformation to each element of the input.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/transform/unary_transform_basic.py
           :language: python
           :start-after: # example-begin

        When working with custom struct types, you need to provide type annotations
        to help with type inference. See the binary transform struct example for reference:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/struct/struct_transform.py
           :language: python
           :start-after: # example-begin


    Args:
        d_in: Device array or iterator containing the input sequence of data items.
        d_out: Device array or iterator to store the result of the transformation.
        op: Unary operation to apply to each element.
            The signature is ``(T) -> U``, where ``T`` is
            the input data type and ``U`` is the output data type.
            Can reference device arrays as globals/closures - they will be automatically captured.
        num_items: Number of items to transform.
        stream: CUDA stream to use for the operation.
    """
    op_adapter = make_op_adapter(op)
    transformer = make_unary_transform(d_in, d_out, op_adapter)
    transformer(d_in, d_out, op_adapter, num_items, stream)


def binary_transform(
    d_in1: DeviceArrayLike | IteratorT,
    d_in2: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    op: Operator,
    num_items: int,
    stream=None,
):
    """
    Performs device-wide binary transform.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``binary_transform`` is used to apply a transformation to pairs of elements from two input sequences.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/transform/binary_transform_basic.py
           :language: python
           :start-after: # example-begin

        When working with custom struct types, you need to provide type annotations
        to help with type inference. See the following example:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/struct/struct_transform.py
           :language: python
           :start-after: # example-begin


    Args:
        d_in1: Device array or iterator containing the first input sequence of data items.
        d_in2: Device array or iterator containing the second input sequence of data items.
        d_out: Device array or iterator to store the result of the transformation.
        op: Binary operation.
            The signature is ``(T1, T2) -> U``, where ``T1`` and ``T2`` are the input data types and
            ``U`` is the output data type.
            Can reference device arrays as globals/closures - they will be automatically captured.
        num_items: Number of items to transform.
        stream: CUDA stream to use for the operation.
    """
    op_adapter = make_op_adapter(op)
    transformer = make_binary_transform(d_in1, d_in2, d_out, op_adapter)
    transformer(d_in1, d_in2, d_out, op_adapter, num_items, stream)



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""ConstantIterator implementation."""

from __future__ import annotations

from textwrap import dedent

import numpy as np

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir, cpp_type_from_descriptor
from ..types import from_numpy_dtype
from ._base import IteratorBase
from ._common import CUDA_PREAMBLE


class ConstantIterator(IteratorBase):
    """
    Iterator representing a sequence of constant values.

    Similar to `thrust::constant_iterator <https://nvidia.github.io/cccl/thrust/api/classthrust_1_1constant__iterator.html>`_.

    Every dereference returns the same constant value.

    Example:
        The code snippet below demonstrates the usage of a ``ConstantIterator``
        representing a sequence of constant values:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/constant_iterator_basic.py
            :language: python
            :start-after: # example-begin

    Args:
        value: The value of every item in the sequence
    """

    def __init__(self, value: np.number):
        """
        Create a constant iterator with the given value.

        Args:
            value: The constant value (must be a numpy scalar)
        """
        if not isinstance(value, np.generic):
            value = np.array(value).flatten()[0]

        self._constant_value = value
        value_type = from_numpy_dtype(value.dtype)
        state_bytes = value.tobytes()

        super().__init__(
            state_bytes=state_bytes,
            state_alignment=value_type.alignment,
            value_type=value_type,
        )

    def _make_advance_op(self) -> Op:
        symbol = self._make_advance_symbol()

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {symbol}(void*, void*) {{
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def _make_input_deref_op(self) -> Op | None:
        symbol = self._make_input_deref_symbol()
        cpp_type = cpp_type_from_descriptor(self._value_type)

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {symbol}(void* state, void* result) {{
                *static_cast<{cpp_type}*>(result) = *static_cast<{cpp_type}*>(state);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)
        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[],
        )

    def _make_output_deref_op(self) -> Op | None:
        return None

    def __add__(self, offset: int) -> "ConstantIterator":
        """Return a new ConstantIterator (value doesn't change with position)."""
        return ConstantIterator(self._constant_value)



# Copyright (c) 2024-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Callable, Protocol, TypeVar

from .iterators import IteratorBase
from .op import OpAdapter, OpKind, RawOp
from .struct import _Struct


class DeviceArrayLike(Protocol):
    """Protocol for array-like objects that expose device memory via CUDA Array Interface.

    Any object implementing the ``__cuda_array_interface__`` attribute can be used
    where a :class:`DeviceArrayLike` is expected. This includes CuPy arrays, Numba
    device arrays, PyTorch CUDA tensors, and other GPU array types.

    See `CUDA Array Interface specification <https://nvidia.github.io/numba-cuda/user/cuda_array_interface.html>`_
    for details.
    """

    __cuda_array_interface__: dict


class StreamLike(Protocol):
    """Protocol for CUDA stream objects.

    Any object implementing the ``__cuda_stream__()`` method can be used where a
    :class:`StreamLike` is expected. This includes stream objects from CuPy,
    Numba CUDA, PyTorch, and other CUDA libraries.
    """

    def __cuda_stream__(self) -> tuple[int, int]: ...


GpuStruct = TypeVar("GpuStruct", bound=_Struct)
"""
Instance of types created with :class:`cuda.compute.struct.gpu_struct`.
"""

IteratorT = TypeVar("IteratorT", bound=IteratorBase)
"""Type variable for iterator objects.

Represents any subclass of :class:`IteratorBase <cuda.compute.iterators.IteratorBase>`.
See :py:mod:`cuda.compute.iterators` for all available iterators.
"""

Operator = Callable | OpKind | RawOp | OpAdapter
"""Type alias for operator objects passed to algorithm functions.

Algorithms accept the following objects as operators:

* Python functions or lambdas implementing the operator. This function will be JIT
  compiled into device code using `numba.cuda <https://nvidia.github.io/numba-cuda/>`_.

* :class:`OpKind <cuda.compute.op.OpKind>` enumerators which are pre-defined constants
  for common operations.

* :class:`RawOp <cuda.compute.op.RawOp` objects containing pre-compiled device code.
"""

__all__ = ["DeviceArrayLike", "GpuStruct", "IteratorT", "Operator"]



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""ShuffleIterator implementation."""

from __future__ import annotations

import struct
from textwrap import dedent

import numpy as np

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir
from ..types import from_numpy_dtype
from ._base import IteratorBase
from ._common import CUDA_PREAMBLE

_SHUFFLE_STATE_STRUCT = """\
struct ShuffleState {
    int64_t current_index;
    uint64_t num_items;
    uint64_t seed;
};"""


class ShuffleIterator(IteratorBase):
    """
    Iterator that produces a deterministic random permutation of indices.

    At position ``i``, yields ``bijection(i)`` where the bijection is a random
    permutation of ``[0, num_items)`` parameterized by ``seed``.

    Example:
        The code snippet below demonstrates the usage of a ``ShuffleIterator``
        to randomly permute indices:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/shuffle_iterator_basic.py
            :language: python
            :start-after: # example-begin

    Args:
        num_items: Number of elements in the domain to permute. Must be > 0.
        seed: Seed for the random permutation. Different seeds produce
            different (deterministic) permutations. Defaults to 0.
    """

    __slots__ = ["_num_items", "_seed", "_current_index"]

    def __init__(self, num_items: int, seed: int = 0, *, _current_index: int = 0):
        if num_items <= 0:
            raise ValueError("num_items must be > 0")

        self._num_items = int(num_items)
        self._seed = int(seed)
        self._current_index = int(_current_index)

        # State layout matches C++ ShuffleState:
        #   int64_t  current_index  (offset 0,  size 8)
        #   uint64_t num_items      (offset 8,  size 8)
        #   uint64_t seed           (offset 16, size 8)
        state_bytes = struct.pack(
            "<qQQ", self._current_index, self._num_items, self._seed
        )

        super().__init__(
            state_bytes=state_bytes,
            state_alignment=8,
            value_type=from_numpy_dtype(np.dtype("int64")),
        )

    def _make_advance_op(self) -> Op:
        symbol = self._make_advance_symbol()

        source = dedent(f"""
            {CUDA_PREAMBLE}

            {_SHUFFLE_STATE_STRUCT}

            extern "C" __device__ void {symbol}(void* state, void* offset) {{
                auto s = static_cast<ShuffleState*>(state);
                auto dist = *static_cast<int64_t*>(offset);
                s->current_index += dist;
            }}
        """)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=compile_cpp_to_ltoir(source),
            extra_ltoirs=[],
        )

    def _make_input_deref_op(self) -> Op | None:
        symbol = self._make_input_deref_symbol()

        # Note: a potential optimization is to avoid constructing
        # `cuda::random_bijection` objects upon every dereference,
        # instead constructing it once and using it as the state
        # object. The tradeoff is that it would require a C++
        # extension providing a constructor for
        # `cuda::random_bijection` objects, since we would now be
        # doing it on the host.  See discussion in #7721.
        source = dedent(f"""
            #include <cuda/__random/random_bijection.h>
            #include <cuda/__random/pcg_engine.h>
            {CUDA_PREAMBLE}

            {_SHUFFLE_STATE_STRUCT}

            // __noinline__ is required to prevent the compiler from merging
            // this function's register usage into the calling kernel during LTO
            // inlining. feistel_bijection constructs 24 round keys with
            // UNROLL_FULL, which exhausts the kernel's register budget and
            // causes spilling to local memory (LDL/STL instructions).
            // Keeping it non-inlined gives it an isolated register frame.
            __device__ __noinline__ int64_t __shuffle_apply(uint64_t num_items, uint64_t seed, uint64_t idx) {{
                cuda::pcg64 rng(seed);
                cuda::random_bijection<uint64_t> bijection(num_items, rng);
                return static_cast<int64_t>(bijection(idx));
            }}

            extern "C" __device__ void {symbol}(void* state, void* result) {{
                const auto* s = static_cast<const ShuffleState*>(state);
                *static_cast<int64_t*>(result) = __shuffle_apply(
                    s->num_items, s->seed, static_cast<uint64_t>(s->current_index));
            }}
        """)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=compile_cpp_to_ltoir(source),
            extra_ltoirs=[],
        )

    def _make_output_deref_op(self) -> Op | None:
        return None

    def __add__(self, offset: int) -> "ShuffleIterator":
        return ShuffleIterator(
            self._num_items,
            self._seed,
            _current_index=self._current_index + offset,
        )



# Copyright (c) 2024-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

import numpy as np

from ... import _bindings
from ... import _cccl_interop as cccl
from ..._caching import cache_with_registered_key_functions
from ..._cccl_interop import call_build, set_cccl_iterator_state
from ..._utils.protocols import (
    get_data_pointer,
    validate_and_get_stream,
)
from ..._utils.temp_storage_buffer import TempStorageBuffer
from ...typing import DeviceArrayLike
from ._sort_common import DoubleBuffer, SortOrder, _get_arrays


class _SegmentedSort:
    __slots__ = [
        "build_result",
        "d_in_keys_cccl",
        "d_out_keys_cccl",
        "d_in_values_cccl",
        "d_out_values_cccl",
        "start_offsets_in_cccl",
        "end_offsets_in_cccl",
    ]

    def __init__(
        self,
        d_in_keys: DeviceArrayLike | DoubleBuffer,
        d_out_keys: DeviceArrayLike | None,
        d_in_values: DeviceArrayLike | DoubleBuffer | None,
        d_out_values: DeviceArrayLike | None,
        start_offsets_in: DeviceArrayLike,
        end_offsets_in: DeviceArrayLike,
        order: SortOrder,
    ):
        d_in_keys_array, d_out_keys_array, d_in_values_array, d_out_values_array = (
            _get_arrays(d_in_keys, d_out_keys, d_in_values, d_out_values)
        )

        self.d_in_keys_cccl = cccl.to_cccl_input_iter(d_in_keys_array)
        self.d_out_keys_cccl = cccl.to_cccl_output_iter(d_out_keys_array)
        self.d_in_values_cccl = cccl.to_cccl_input_iter(d_in_values_array)
        self.d_out_values_cccl = cccl.to_cccl_output_iter(d_out_values_array)
        self.start_offsets_in_cccl = cccl.to_cccl_input_iter(start_offsets_in)
        self.end_offsets_in_cccl = cccl.to_cccl_input_iter(end_offsets_in)

        self.build_result = call_build(
            _bindings.DeviceSegmentedSortBuildResult,
            _bindings.SortOrder.ASCENDING
            if order is SortOrder.ASCENDING
            else _bindings.SortOrder.DESCENDING,
            self.d_in_keys_cccl,
            self.d_in_values_cccl,
            self.start_offsets_in_cccl,
            self.end_offsets_in_cccl,
        )

    def __call__(
        self,
        temp_storage,
        d_in_keys,
        d_out_keys,
        d_in_values,
        d_out_values,
        num_items,
        num_segments,
        start_offsets_in,
        end_offsets_in,
        stream=None,
    ):
        if num_segments > np.iinfo(np.int32).max:
            raise RuntimeError(
                "Segmented sort does not currently support more than 2^31-1 segments."
            )
        d_in_keys_array, d_out_keys_array, d_in_values_array, d_out_values_array = (
            _get_arrays(d_in_keys, d_out_keys, d_in_values, d_out_values)
        )

        set_cccl_iterator_state(self.d_in_keys_cccl, d_in_keys_array)
        set_cccl_iterator_state(self.d_out_keys_cccl, d_out_keys_array)
        if d_in_values_array is not None:
            set_cccl_iterator_state(self.d_in_values_cccl, d_in_values_array)
        if d_out_values_array is not None:
            set_cccl_iterator_state(self.d_out_values_cccl, d_out_values_array)
        set_cccl_iterator_state(self.start_offsets_in_cccl, start_offsets_in)
        set_cccl_iterator_state(self.end_offsets_in_cccl, end_offsets_in)

        stream_handle = validate_and_get_stream(stream)
        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            d_temp_storage = get_data_pointer(temp_storage)

        # Detect overwrite mode and selector, similar to radix sort
        is_overwrite_okay = isinstance(d_in_keys, DoubleBuffer)
        selector = -1

        temp_storage_bytes, selector = self.build_result.compute(
            d_temp_storage,
            temp_storage_bytes,
            self.d_in_keys_cccl,
            self.d_out_keys_cccl,
            self.d_in_values_cccl,
            self.d_out_values_cccl,
            num_items,
            num_segments,
            self.start_offsets_in_cccl,
            self.end_offsets_in_cccl,
            is_overwrite_okay,
            selector,
            stream_handle,
        )

        if is_overwrite_okay and temp_storage is not None:
            assert selector in (0, 1)
            assert isinstance(d_in_keys, DoubleBuffer)
            d_in_keys.selector = selector
            if d_in_values is not None:
                assert isinstance(d_in_values, DoubleBuffer)
                d_in_values.selector = selector

        return temp_storage_bytes


@cache_with_registered_key_functions
def make_segmented_sort(
    d_in_keys: DeviceArrayLike | DoubleBuffer,
    d_out_keys: DeviceArrayLike | None,
    d_in_values: DeviceArrayLike | DoubleBuffer | None,
    d_out_values: DeviceArrayLike | None,
    start_offsets_in: DeviceArrayLike,
    end_offsets_in: DeviceArrayLike,
    order: SortOrder,
):
    """
    Performs a device-wide segmented sort using the specified keys and values.

    Example:
        Below, ``make_segmented_sort`` is used to create a segmented sort object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/sort/segmented_sort_object.py
            :language: python
            :start-after: # example-begin

    Args:
        d_in_keys: Device array or DoubleBuffer containing the input keys to be sorted
        d_out_keys: Device array to store the sorted keys
        d_in_values: Optional Device array or DoubleBuffer containing the input values to be sorted
        d_out_values: Device array to store the sorted values
        start_offsets_in: Device array or iterator containing the sequence of beginning offsets
        end_offsets_in: Device array or iterator containing the sequence of ending offsets
        order: SortOrder specifying the order of the sort

    Returns:
        A callable object that can be used to perform the segmented sort
    """
    return _SegmentedSort(
        d_in_keys,
        d_out_keys,
        d_in_values,
        d_out_values,
        start_offsets_in,
        end_offsets_in,
        order,
    )


def segmented_sort(
    d_in_keys: DeviceArrayLike | DoubleBuffer,
    d_out_keys: DeviceArrayLike | None,
    d_in_values: DeviceArrayLike | DoubleBuffer | None,
    d_out_values: DeviceArrayLike | None,
    num_items: int,
    num_segments: int,
    start_offsets_in: DeviceArrayLike,
    end_offsets_in: DeviceArrayLike,
    order: SortOrder,
    stream=None,
):
    """
    Performs device-wide segmented sort.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``segmented_sort`` is used to perform a segmented sort. It also rearranges the values according to the keys' order.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/sort/segmented_sort_basic.py
            :language: python
            :start-after: # example-begin


        In the following example, ``segmented_sort`` is used to perform a segmented sort with a ``DoubleBuffer`` for reduced temporary storage.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/sort/segmented_sort_buffer.py
            :language: python
            :start-after: # example-begin

    Args:
        d_in_keys: Device array or DoubleBuffer containing the input keys to be sorted
        d_out_keys: Device array to store the sorted keys (optional)
        d_in_values: Device array or DoubleBuffer containing the input values to be sorted (optional)
        d_out_values: Device array to store the sorted values (optional)
        num_items: Total number of items to sort
        num_segments: Number of segments to sort
        start_offsets_in: Device array or iterator containing the sequence of beginning offsets
        end_offsets_in: Device array or iterator containing the sequence of ending offsets
        order: Sort order (ascending or descending)
        stream: CUDA stream for the operation (optional)
    """
    sorter = make_segmented_sort(
        d_in_keys,
        d_out_keys,
        d_in_values,
        d_out_values,
        start_offsets_in,
        end_offsets_in,
        order,
    )
    tmp_storage_bytes = sorter(
        None,
        d_in_keys,
        d_out_keys,
        d_in_values,
        d_out_values,
        num_items,
        num_segments,
        start_offsets_in,
        end_offsets_in,
        stream,
    )
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    sorter(
        tmp_storage,
        d_in_keys,
        d_out_keys,
        d_in_values,
        d_out_values,
        num_items,
        num_segments,
        start_offsets_in,
        end_offsets_in,
        stream,
    )



# Copyright (c) 2024-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


from __future__ import annotations

from .. import _bindings, types
from .. import _cccl_interop as cccl
from .._caching import cache_with_registered_key_functions
from .._cccl_interop import call_build, set_cccl_iterator_state
from .._utils.protocols import (
    get_data_pointer,
    validate_and_get_stream,
)
from .._utils.temp_storage_buffer import TempStorageBuffer
from ..op import OpAdapter, make_op_adapter
from ..typing import DeviceArrayLike, IteratorT, Operator


class _UniqueByKey:
    __slots__ = [
        "build_result",
        "d_in_keys_cccl",
        "d_in_items_cccl",
        "d_out_keys_cccl",
        "d_out_items_cccl",
        "d_out_num_selected_cccl",
        "op_cccl",
    ]

    def __init__(
        self,
        d_in_keys: DeviceArrayLike | IteratorT,
        d_in_items: DeviceArrayLike | IteratorT,
        d_out_keys: DeviceArrayLike | IteratorT,
        d_out_items: DeviceArrayLike | IteratorT,
        d_out_num_selected: DeviceArrayLike,
        op: OpAdapter,
    ):
        self.d_in_keys_cccl = cccl.to_cccl_input_iter(d_in_keys)
        self.d_in_items_cccl = cccl.to_cccl_input_iter(d_in_items)
        self.d_out_keys_cccl = cccl.to_cccl_output_iter(d_out_keys)
        self.d_out_items_cccl = cccl.to_cccl_output_iter(d_out_items)
        self.d_out_num_selected_cccl = cccl.to_cccl_output_iter(d_out_num_selected)

        # Compile the op - unique_by_key expects bool return (comparison)
        value_type = cccl.get_value_type(d_in_keys)
        self.op_cccl = op.compile((value_type, value_type), types.uint8)

        self.build_result = call_build(
            _bindings.DeviceUniqueByKeyBuildResult,
            self.d_in_keys_cccl,
            self.d_in_items_cccl,
            self.d_out_keys_cccl,
            self.d_out_items_cccl,
            self.d_out_num_selected_cccl,
            self.op_cccl,
        )

    def __call__(
        self,
        temp_storage,
        d_in_keys: DeviceArrayLike | IteratorT,
        d_in_items: DeviceArrayLike | IteratorT,
        d_out_keys: DeviceArrayLike | IteratorT,
        d_out_items: DeviceArrayLike | IteratorT,
        d_out_num_selected: DeviceArrayLike,
        op: Operator,
        num_items: int,
        stream=None,
    ):
        set_cccl_iterator_state(self.d_in_keys_cccl, d_in_keys)
        set_cccl_iterator_state(self.d_in_items_cccl, d_in_items)
        set_cccl_iterator_state(self.d_out_keys_cccl, d_out_keys)
        set_cccl_iterator_state(self.d_out_items_cccl, d_out_items)
        set_cccl_iterator_state(self.d_out_num_selected_cccl, d_out_num_selected)

        # Update op state for stateful ops
        op_adapter = make_op_adapter(op)
        self.op_cccl.state = op_adapter.get_state()

        stream_handle = validate_and_get_stream(stream)
        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            # Note: this is slightly slower, but supports all ndarray-like objects as long as they support CAI
            # TODO: switch to use gpumemoryview once it's ready
            d_temp_storage = get_data_pointer(temp_storage)

        temp_storage_bytes = self.build_result.compute(
            d_temp_storage,
            temp_storage_bytes,
            self.d_in_keys_cccl,
            self.d_in_items_cccl,
            self.d_out_keys_cccl,
            self.d_out_items_cccl,
            self.d_out_num_selected_cccl,
            self.op_cccl,
            num_items,
            stream_handle,
        )
        return temp_storage_bytes


@cache_with_registered_key_functions
def make_unique_by_key(
    d_in_keys: DeviceArrayLike | IteratorT,
    d_in_items: DeviceArrayLike | IteratorT,
    d_out_keys: DeviceArrayLike | IteratorT,
    d_out_items: DeviceArrayLike | IteratorT,
    d_out_num_selected: DeviceArrayLike,
    op: Operator,
):
    """Implements a device-wide unique by key operation using ``d_in_keys`` and the comparison operator ``op``. Only the first key and its value from each run is selected and the total number of items selected is also reported.

    Example:
        Below, ``make_unique_by_key`` is used to create a unique by key object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/unique/unique_by_key_object.py
          :language: python
          :start-after: # example-begin


    Args:
        d_in_keys: Device array or iterator containing the input sequence of keys
        d_in_items: Device array or iterator that contains each key's corresponding item
        d_out_keys: Device array or iterator to store the outputted keys
        d_out_items: Device array or iterator to store each outputted key's item
        d_out_num_selected: Device array to store how many items were selected
        op: Callable or OpKind representing the equality operator

    Returns:
        A callable object that can be used to perform unique by key
    """
    op_adapter = make_op_adapter(op)
    return _UniqueByKey(
        d_in_keys, d_in_items, d_out_keys, d_out_items, d_out_num_selected, op_adapter
    )


def unique_by_key(
    d_in_keys: DeviceArrayLike | IteratorT,
    d_in_items: DeviceArrayLike | IteratorT,
    d_out_keys: DeviceArrayLike | IteratorT,
    d_out_items: DeviceArrayLike | IteratorT,
    d_out_num_selected: DeviceArrayLike,
    op: Operator,
    num_items: int,
    stream=None,
):
    """
    Performs device-wide unique by key operation using the single-phase API.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``unique_by_key`` is used to populate the arrays of output keys and items with the first key and its corresponding item from each sequence of equal keys. It also outputs the number of items selected.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/unique/unique_by_key_basic.py
            :language: python
            :start-after: # example-begin


    Args:
        d_in_keys: Device array or iterator containing the input sequence of keys
        d_in_items: Device array or iterator that contains each key's corresponding item
        d_out_keys: Device array or iterator to store the outputted keys
        d_out_items: Device array or iterator to store each outputted key's item
        d_out_num_selected: Device array to store how many items were selected
        op: Callable or OpKind representing the equality operator
        num_items: Number of items to process
        stream: CUDA stream for the operation (optional)
    """
    uniquer = make_unique_by_key(
        d_in_keys, d_in_items, d_out_keys, d_out_items, d_out_num_selected, op
    )
    tmp_storage_bytes = uniquer(
        None,
        d_in_keys,
        d_in_items,
        d_out_keys,
        d_out_items,
        d_out_num_selected,
        op,
        num_items,
        stream,
    )
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    uniquer(
        tmp_storage,
        d_in_keys,
        d_in_items,
        d_out_keys,
        d_out_items,
        d_out_num_selected,
        op,
        num_items,
        stream,
    )



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""
cuda.coop.block_exchange
====================================

This module provides a set of :ref:`collective <collective-primitives>` methods
for rearranging data partitioned across CUDA thread blocks.

Supported C++ APIs
++++++++++++++++++

The following :cpp:class:`cub.BlockExchange` APIs are supported:

    StripedToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD])
    StripedToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], ::cuda::std::false_type)
    StripedToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], ::cuda::std::true_type)
    StripedToBlocked void (T (&)[ITEMS_PER_THREAD])

Unsupported C++ APIs
++++++++++++++++++++

The following :cpp:class:`cub.BlockExchange` APIs are not yet supported:

    BlockedToStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD])
    BlockedToStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], ::cuda::std::false_type)
    BlockedToStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], ::cuda::std::true_type)
    BlockedToStriped void (T (&)[ITEMS_PER_THREAD])

    BlockedToWarpStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD])
    BlockedToWarpStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], ::cuda::std::false_type)
    BlockedToWarpStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], ::cuda::std::true_type)
    BlockedToWarpStriped void (T (&)[ITEMS_PER_THREAD])

    ScatterToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD])
    ScatterToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD], ::cuda::std::false_type)
    ScatterToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD], ::cuda::std::true_type)
    ScatterToBlocked template void (T (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD])

    ScatterToStripedFlagged template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD], ValidFlag (&)[ITEMS_PER_THREAD])
    ScatterToStripedFlagged template void (T (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD], ValidFlag (&)[ITEMS_PER_THREAD])

    ScatterToStripedGuarded template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD])
    ScatterToStripedGuarded template void (T (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD])

    ScatterToStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD])
    ScatterToStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD], ::cuda::std::false_type)
    ScatterToStriped template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD], ::cuda::std::true_type)
    ScatterToStriped template void (T (&)[ITEMS_PER_THREAD], OffsetT (&)[ITEMS_PER_THREAD])

    WarpStripedToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD])
    WarpStripedToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], ::cuda::std::false_type)
    WarpStripedToBlocked template void (const T (&)[ITEMS_PER_THREAD], OutputT (&)[ITEMS_PER_THREAD], ::cuda::std::true_type)
    WarpStripedToBlocked void (T (&)[ITEMS_PER_THREAD])
"""

from enum import IntEnum, auto

import numba

from .._common import (
    make_binary_tempfile,
    normalize_dim_param,
    normalize_dtype_param,
)
from .._types import (
    Algorithm,
    Dependency,
    DependentArray,
    Invocable,
    Pointer,
    TemplateParameter,
    numba_type_to_wrapper,
)
from .._typing import (
    DimType,
    DtypeType,
)


class BlockExchangeType(IntEnum):
    """
    Enum representing the type of block exchange operation.  Currently
    only :py:attr:`StripedToBlocked` is supported.
    """

    StripedToBlocked = auto()


def make_exchange(
    block_exchange_type: BlockExchangeType,
    dtype: DtypeType,
    threads_per_block: DimType,
    items_per_thread: int,
    warp_time_slicing: bool = False,
    methods: dict = None,
):
    """
    Creates a block-wide exchange primitive for rearranging data
    partitioned across CUDA thread blocks.

    Example:
        The snippet below shows how to create and invoke the returned
        ``block_exchange`` primitive for striped-to-blocked exchange.

        .. code-block:: python

           block_exchange = coop.block.make_exchange(
               coop.block.BlockExchangeType.StripedToBlocked,
               dtype=numba.int32,
               threads_per_block=128,
               items_per_thread=4,
           )
           temp_storage_bytes = block_exchange.temp_storage_bytes

           @cuda.jit(link=block_exchange.files)
           def kernel(thread_data):
               temp_storage = cuda.shared.array(
                   shape=temp_storage_bytes,
                   dtype=numba.uint8,
               )
               block_exchange(temp_storage, thread_data)

    :param block_exchange_type: Exchange mode to perform. Currently, only
        :py:attr:`StripedToBlocked` is supported.
    :param dtype: Data type of input and output values.
    :type dtype: :py:class:`cuda.coop._typing.DtypeType`
    :param threads_per_block: Number of threads in the block.
    :type threads_per_block: :py:class:`cuda.coop._typing.DimType`
    :param items_per_thread: Number of items owned by each thread.
    :type items_per_thread: int
    :param warp_time_slicing: Whether to use warp time-slicing. If true,
        shared memory usage is reduced at the expense of parallelism.
    :type warp_time_slicing: bool, optional

    :param methods: Optional method dictionary for user-defined types.
    :type methods: dict, optional

    :raises ValueError: If ``block_exchange_type`` is not a valid enum
        value of :py:class:`BlockExchangeType`.

    :raises ValueError: If ``items_per_thread`` is less than 1.

    :returns: An :py:class:`cuda.coop._types.Invocable` object
        representing the specialized kernel callable from a Numba JIT'd
        CUDA kernel.

    """
    # Validate initial parameters.
    if block_exchange_type not in BlockExchangeType:
        raise ValueError(
            "block_exchange_type must be a valid BlockExchangeType enum "
            f"value; got: {block_exchange_type!r}"
        )
    elif block_exchange_type != BlockExchangeType.StripedToBlocked:
        raise ValueError(
            "block_exchange_type must be BlockExchangeType.StripedToBlocked; "
            f"got: {block_exchange_type!r}"
        )

    if items_per_thread < 1:
        raise ValueError("items_per_thread must be greater than or equal to 1")

    # Normalize parameters.
    dim = normalize_dim_param(threads_per_block)
    dtype = normalize_dtype_param(dtype)

    specialization_kwds = {
        "T": dtype,
        "BLOCK_DIM_X": dim[0],
        "ITEMS_PER_THREAD": items_per_thread,
        "WARP_TIME_SLICING": int(warp_time_slicing),
        "BLOCK_DIM_Y": dim[1],
        "BLOCK_DIM_Z": dim[2],
    }

    template_parameters = [
        TemplateParameter("T"),
        TemplateParameter("BLOCK_DIM_X"),
        TemplateParameter("ITEMS_PER_THREAD"),
        TemplateParameter("WARP_TIME_SLICING"),
        TemplateParameter("BLOCK_DIM_Y"),
        TemplateParameter("BLOCK_DIM_Z"),
    ]

    # In other modules, like block scan, items_per_thread affects the calling
    # convention the user needs to adopt in their kernel, as it pertains to
    # input and return parameters, e.g., with items_per_thread = 1, you'd use
    #   output = block_scan(temp_storage, input)
    # Whereas with items_per_thread > 1, you'd use:
    #   block_scan(temp_storage, input, output)
    # This idiom does not apply for block exchange, so we don't specialize
    # the parameters based on items_per_thread.
    parameters = [
        # Signature:
        # void BlockExchange<T, BLOCK_DIM_X, ITEMS_PER_THREAD,
        #                    WARP_TIME_SLICING, BLOCK_DIM_Y, BLOCK_DIM_Z>(
        #     temp_storage
        # )::StripedToBlocked(
        #   const T (&)[ITEMS_PER_THREAD] input_items,
        #   OutputT (&)[ITEMS_PER_THREAD] output_items,
        # )
        [
            # temp_storage
            Pointer(numba.uint8),
            # T (&)[ITEMS_PER_THREAD] items
            DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
        ],
        # Signature:
        # void BlockExchange<T, BLOCK_DIM_X, ITEMS_PER_THREAD,
        #                    WARP_TIME_SLICING, BLOCK_DIM_Y, BLOCK_DIM_Z>(
        #     temp_storage
        # )::StripedToBlocked(
        #   const T (&)[ITEMS_PER_THREAD] input_items,
        #   OutputT (&)[ITEMS_PER_THREAD] output_items,
        # )
        [
            # temp_storage
            Pointer(numba.uint8),
            # const T (&)[ITEMS_PER_THREAD] input_items
            DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
            # OutputT (&)[ITEMS_PER_THREAD] output_items
            DependentArray(Dependency("T"), Dependency("ITEMS_PER_THREAD")),
        ],
    ]

    # If we have a non-None `methods`, we're dealing with user-defined types.
    if methods is not None:
        type_definitions = [
            numba_type_to_wrapper(dtype, methods=methods),
        ]
    else:
        type_definitions = None

    template = Algorithm(
        "BlockExchange",
        "StripedToBlocked",
        "block_exchange",
        ["cub/block/block_exchange.cuh"],
        template_parameters,
        parameters,
        fake_return=False,
        type_definitions=type_definitions,
    )

    specialization = template.specialize(specialization_kwds)
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir()
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""TransformIterator implementation."""

from __future__ import annotations

from textwrap import dedent

from .._bindings import Op, OpKind
from .._cpp_compile import compile_cpp_to_ltoir, make_variable_declaration
from ..op import make_op_adapter
from ..types import TypeDescriptor, signature_from_annotations
from ._base import IteratorBase
from ._common import CUDA_PREAMBLE, ensure_iterator


class TransformIterator(IteratorBase):
    """
    An iterator that applies a unary function to elements as they are read from an underlying iterator.

    Similar to `thrust::transform_iterator <https://nvidia.github.io/cccl/thrust/api/classthrust_1_1transform__iterator.html>`_.

    For input iteration (default): reads from underlying, applies transform, returns result.
    For output iteration: applies transform to input values, writes to underlying.

    Example:
        The code snippet below demonstrates the usage of a ``TransformIterator`` composed with a ``CountingIterator``
        to transform the input before performing a reduction:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/transform_iterator_basic.py
            :language: python
            :start-after: # example-begin

    Args:
        underlying: The underlying iterator or device array
        transform_op: The unary operation to apply
        output_value_type: TypeDescriptor for the output type (optional, will be inferred if not provided)
        is_input: True for input iterator (default), False for output iterator
    """

    __slots__ = [
        "_underlying",
        "_transform_op",
        "_value_type",
        "_is_input",
        "_compiled_op",
    ]

    def __init__(
        self,
        underlying,
        transform_op,
        value_type: TypeDescriptor | None = None,
        is_input: bool = True,
    ):
        """
        Create a transform iterator.

        Args:
            underlying: The underlying iterator or device array to transform
            transform_op: The unary transform operation (callable or OpKind)
            value_type: TypeDescriptor for the transformed value type.
                            For input iterators: inferred if None.
                            For output iterators: must be provided or have annotations.
            is_input: True for input iterator, False for output iterator
        """
        underlying = ensure_iterator(underlying)

        self._underlying = underlying
        self._transform_op = make_op_adapter(transform_op)
        self._is_input = is_input
        self._compiled_op = None  # Lazy compiled Op

        # Determine value type
        if value_type is None:
            if is_input:
                # value_type is the return type
                _, value_type = signature_from_annotations(transform_op)
                if value_type is None:
                    value_type = self._transform_op.get_return_type(
                        (underlying.value_type,)
                    )
            else:
                # value_type is the input type
                input_types, _ = signature_from_annotations(transform_op)
                if len(input_types) != 1:
                    raise ValueError(
                        "TransformOutputIterator transform function must take exactly one argument with type annotation"
                    )
                value_type = input_types[0]
        assert value_type is not None
        self._value_type = value_type

        super().__init__(
            state_bytes=bytes(self._underlying.state),
            state_alignment=self._underlying.state_alignment,
            value_type=value_type,
        )

    def _get_compiled_op(self):
        """Get the compiled Op, compiling lazily if needed."""
        if self._compiled_op is None:
            if self._is_input:
                input_type = self._underlying.value_type
                output_type = self._value_type
            else:
                input_type = self._value_type
                output_type = self._underlying.value_type

            self._compiled_op = self._transform_op.compile(
                (input_type,),
                output_type,
            )
        return self._compiled_op

    def _make_advance_op(self) -> Op:
        """Provide Op for advance that delegates to underlying iterator."""
        child_op = self._underlying.get_advance_op()
        symbol = self._make_advance_symbol()

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {child_op.name}(void* state, void* offset);

            extern "C" __device__ void {symbol}(void* state, void* offset) {{
                {child_op.name}(state, offset);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[child_op.ltoir, *child_op.extra_ltoirs],
        )

    def _make_input_deref_op(self) -> Op | None:
        """Provide Op for input dereference that reads from underlying then transforms."""
        if not self._is_input:
            return None

        child_op = self._underlying.get_input_deref_op()
        if child_op is None:
            raise ValueError("Underlying iterator must support input dereference")

        compiled_op = self._get_compiled_op()
        symbol = self._make_input_deref_symbol()
        temp_decl = make_variable_declaration(self._underlying.value_type, "temp")

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {child_op.name}(void* state, void* result);
            extern "C" __device__ void {compiled_op.name}(void* input, void* output);

            extern "C" __device__ void {symbol}(void* state, void* result) {{
                {temp_decl}
                {child_op.name}(state, &temp);
                {compiled_op.name}(&temp, result);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[
                compiled_op.ltoir,
                *compiled_op.extra_ltoirs,
                child_op.ltoir,
                *child_op.extra_ltoirs,
            ],
        )

    def _make_output_deref_op(self) -> Op | None:
        """Provide Op for output dereference that transforms then writes to underlying."""
        if self._is_input:
            return None

        child_op = self._underlying.get_output_deref_op()
        if child_op is None:
            raise ValueError("Underlying iterator must support output dereference")

        compiled_op = self._get_compiled_op()
        symbol = self._make_output_deref_symbol()
        temp_decl = make_variable_declaration(self._underlying.value_type, "temp")

        source = dedent(f"""
            {CUDA_PREAMBLE}

            extern "C" __device__ void {child_op.name}(void* state, void* value);
            extern "C" __device__ void {compiled_op.name}(void* input, void* output);

            extern "C" __device__ void {symbol}(void* state, void* value) {{
                {temp_decl}
                {compiled_op.name}(value, &temp);
                {child_op.name}(state, &temp);
            }}
        """).strip()

        ltoir = compile_cpp_to_ltoir(source)

        return Op(
            operator_type=OpKind.STATELESS,
            name=symbol,
            ltoir=ltoir,
            extra_ltoirs=[
                compiled_op.ltoir,
                *compiled_op.extra_ltoirs,
                child_op.ltoir,
                *child_op.extra_ltoirs,
            ],
        )

    def advance(self, offset: int) -> "TransformIterator":
        """Return a new iterator advanced by offset elements."""
        if not hasattr(self._underlying, "__add__"):
            raise AttributeError("Underlying iterator does not support advance")
        return TransformIterator(
            self._underlying + offset,  # type: ignore[operator, arg-type]
            self._transform_op,
            self._value_type,
            is_input=self._is_input,
        )

    def __add__(self, offset: int) -> "TransformIterator":
        return self.advance(offset)

    def __radd__(self, offset: int) -> "TransformIterator":
        return self.advance(offset)

    @property
    def children(self):
        return (self._underlying,)

    @property
    def kind(self):
        """Return a hashable kind for caching purposes."""
        # Convert _value_type to tuple if it's a list (for output iterators)
        value_type = (
            tuple(self._value_type)
            if isinstance(self._value_type, list)
            else self._value_type
        )
        return (
            "TransformIterator",
            self._is_input,
            self._transform_op,
            self._underlying.kind,
            value_type,
        )


class TransformOutputIterator(TransformIterator):
    """
    An iterator that applies a unary function to values before writing them to an underlying iterator.

    Similar to `thrust::transform_output_iterator <https://nvidia.github.io/cccl/thrust/api/classthrust_1_1transform__output__iterator.html>`_.

    This is a convenience subclass of TransformIterator configured for output mode.

    Example:
        The code snippet below demonstrates the usage of a ``TransformOutputIterator`` to transform the output
        of a reduction before writing to an output array:

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/iterator/transform_output_iterator.py
            :language: python
            :start-after: # example-begin

    Args:
        underlying: The underlying iterator or device array
        transform_op: The operation to be applied to values before they are written
        output_value_type: TypeDescriptor for the input value type (optional, will be extracted from annotations if not provided)
    """

    def __init__(self, underlying, transform_op, output_value_type=None):
        super().__init__(underlying, transform_op, output_value_type, is_input=False)



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from ._bindings import _BINDINGS_AVAILABLE  # type: ignore[attr-defined]

if not _BINDINGS_AVAILABLE:
    __all__ = ["_BINDINGS_AVAILABLE"]

    def __getattr__(name):
        raise AttributeError(
            f"Cannot access 'cuda.compute.{name}' because CUDA bindings are not available."
            "This typically means you're running on a CPU-only machine without CUDA drivers installed."
        )
else:
    from ._caching import clear_all_caches
    from .algorithms import (
        DoubleBuffer,
        SortOrder,
        binary_transform,
        exclusive_scan,
        histogram_even,
        inclusive_scan,
        lower_bound,
        make_binary_transform,
        make_exclusive_scan,
        make_histogram_even,
        make_inclusive_scan,
        make_lower_bound,
        make_merge_sort,
        make_radix_sort,
        make_reduce_into,
        make_segmented_reduce,
        make_segmented_sort,
        make_select,
        make_three_way_partition,
        make_unary_transform,
        make_unique_by_key,
        make_upper_bound,
        merge_sort,
        radix_sort,
        reduce_into,
        segmented_reduce,
        segmented_sort,
        select,
        three_way_partition,
        unary_transform,
        unique_by_key,
        upper_bound,
    )
    from .determinism import Determinism
    from .iterators import (
        CacheModifiedInputIterator,
        ConstantIterator,
        CountingIterator,
        DiscardIterator,
        PermutationIterator,
        ReverseIterator,
        ShuffleIterator,
        TransformIterator,
        TransformOutputIterator,
        ZipIterator,
    )
    from .op import OpKind
    from .struct import gpu_struct

    __all__ = [
        "_BINDINGS_AVAILABLE",
        "binary_transform",
        "clear_all_caches",
        "CacheModifiedInputIterator",
        "ConstantIterator",
        "CountingIterator",
        "DiscardIterator",
        "DoubleBuffer",
        "exclusive_scan",
        "gpu_struct",
        "histogram_even",
        "inclusive_scan",
        "lower_bound",
        "make_binary_transform",
        "make_exclusive_scan",
        "make_select",
        "make_histogram_even",
        "make_inclusive_scan",
        "make_lower_bound",
        "make_merge_sort",
        "make_radix_sort",
        "make_reduce_into",
        "make_segmented_reduce",
        "make_segmented_sort",
        "make_three_way_partition",
        "make_unary_transform",
        "make_unique_by_key",
        "make_upper_bound",
        "merge_sort",
        "OpKind",
        "Determinism",
        "PermutationIterator",
        "radix_sort",
        "reduce_into",
        "ReverseIterator",
        "ShuffleIterator",
        "segmented_reduce",
        "segmented_sort",
        "select",
        "SortOrder",
        "TransformIterator",
        "TransformOutputIterator",
        "three_way_partition",
        "unary_transform",
        "unique_by_key",
        "upper_bound",
        "ZipIterator",
    ]



# Copyright (c) 2025CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import TYPE_CHECKING, Callable, Literal, Tuple, Union

if TYPE_CHECKING:
    import numba
    import numpy as np

    from ._common import dim3

# Type alias for dimension parameters that can be passed to CUDA functions.
DimType = Union["dim3", int, Tuple[int, int], Tuple[int, int, int]]
"""
.. _DimType:

Dimension parameter specification for CUDA functions.

This type alias accepts the following forms:

- A single integer (1D thread/block configuration).
- A tuple of two or three integers representing multi-dimensional
  (2D/3D) CUDA dimensions.
- A ``dim3`` object, which is a namedtuple with three fields: ``x``,
  ``y``, and ``z``.

Examples
--------
.. code-block:: python

    threads = 128  # 1D
    threads = (16, 16)  # 2D
    threads = (8, 8, 4)  # 3D

See Also
--------
- `CUDA dim3 documentation <https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#dim3>`_
"""

DtypeType = Union[str, type, "np.number", "np.dtype", "numba.types.Type"]
"""
.. _DtypeType:

Data type specification for CUDA kernel parameters and operations.

Acceptable forms include:

- Built-in Python numeric types (e.g., ``int``, ``float``).
- NumPy scalar types (e.g., ``np.float32``) or NumPy dtypes (e.g.,
  ``np.dtype('float32')``).
- Numba types (e.g., ``numba.int32``).
- Strings describing the data type (e.g., ``"float32"`` or ``"int64"``).

Examples
--------
.. code-block:: python

    dtype = np.float32
    dtype = "int32"
    dtype = numba.types.float64

See Also
--------
- :mod:`numpy`
- :mod:`numba.types`
"""

# Type alias for scan operators.
ScanOpType = Union[
    # Explicitly named operators.
    Literal[
        "add",
        "plus",
        "mul",
        "multiplies",
        "min",
        "minimum",
        "max",
        "maximum",
        "bit_and",
        "bit_or",
        "bit_xor",
    ],
    # Short-hand operators.
    Literal["+", "*", "&", "|", "^"],
    # Callable objects.
    Callable[["numba.types.Number", "numba.types.Number"], "numba.types.Number"],
    Callable[["np.ndarray", "np.ndarray"], "np.ndarray"],
    Callable[["np.number", "np.number"], "np.number"],
]
"""
.. _ScanOpType:

Specification for binary scan operator used in block-wide CUDA scan operations.

This type can be one of the following:

- **Named operators** (as strings): ``"add"``, ``"mul"``, ``"min"``,
  ``"max"``, ``"bit_and"``, ``"bit_or"``, and ``"bit_xor"``.
- **Symbolic short-hand operators**: ``"+"``, ``"*"``, ``"&"``,
  ``"|"``, and ``"^"``.
- **User-defined callable** that accepts two input values and returns a single
  output value.

"""



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

"""
Base classes for iterators.
"""

from __future__ import annotations

import hashlib
from typing import Hashable

from .._bindings import Iterator, IteratorKind, IteratorState, Op
from .._caching import cache_with_registered_key_functions
from ..types import TypeDescriptor


class IteratorBase:
    """
    Iterators represent streams of data computed on the fly.

    See :py:mod:`cuda.compute.iterators` for available iterators.
    """

    # Subclassing
    # -----------
    #
    # Subclasses must implement the following methods that return
    # Op objects.
    #
    # - _make_advance_op() -> Op
    # - _make_input_deref_op() -> Op | None
    # - _make_output_deref_op() -> Op | None
    #
    # Iterators composed of other iterators must also implement:
    #
    # - children property to return tuple of child iterators for dependency tracking
    #
    # Examples of such "compound" iterators include TransformIterator,
    # PermutationIterator, ReverseIterator and ZipIterator.
    #
    # The base class provides public cached accessors:
    #
    # - get_advance_op() -> Op (cached)
    # - get_input_deref_op() -> Op | None (cached)
    # - get_output_deref_op() -> Op | None (cached)

    __slots__ = [
        "_state_bytes",
        "_state_alignment",
        "_value_type",
        "_advance_op",
        "_input_deref_op",
        "_output_deref_op",
        "_uid_cached",
    ]

    def __init__(
        self,
        state_bytes: bytes,
        state_alignment: int,
        value_type: TypeDescriptor,
    ):
        """
        Args:
            state_bytes: bytes object representing iterator's state
            state_alignment: Alignment of the state
            value_type: Type of dereferenced values
        """
        self._state_bytes = state_bytes
        self._state_alignment = state_alignment
        self._value_type = value_type
        self._advance_op: Op | None = None
        self._input_deref_op: Op | None = None
        self._output_deref_op: Op | None = None
        self._uid_cached: str | None = None

    @property
    def state(self) -> IteratorState:
        """Return the iterator state for CCCL interop."""
        return IteratorState(self._state_bytes)

    @property
    def state_alignment(self) -> int:
        """Return the alignment of the iterator state."""
        return self._state_alignment

    @property
    def value_type(self) -> TypeDescriptor:
        """Return the TypeDescriptor for dereferenced values."""
        return self._value_type

    @property
    def children(self) -> tuple["IteratorBase", ...]:
        """Return child iterators for automatic dependency tracking. Override in subclasses."""
        return ()

    def _get_uid(self) -> str:
        """Return a deterministic unique identifier for this iterator type."""
        if self._uid_cached is None:
            self._uid_cached = _deterministic_suffix(self.kind)
        return self._uid_cached

    def _make_advance_symbol(self) -> str:
        """Generate symbol name for advance operation."""
        return f"{self.__class__.__name__}_advance_{self._get_uid()}"

    def _make_input_deref_symbol(self) -> str:
        """Generate symbol name for input dereference operation."""
        return f"{self.__class__.__name__}_input_deref_{self._get_uid()}"

    def _make_output_deref_symbol(self) -> str:
        """Generate symbol name for output dereference operation."""
        return f"{self.__class__.__name__}_output_deref_{self._get_uid()}"

    def get_advance_op(self) -> Op:
        """Get the cached Op for the advance operation."""
        if self._advance_op is None:
            self._advance_op = self._make_advance_op()
        return self._advance_op

    def get_input_deref_op(self) -> Op | None:
        """Get the cached Op for input dereference operation, or None if not supported."""
        if self._input_deref_op is None:
            self._input_deref_op = self._make_input_deref_op()
        return self._input_deref_op

    def get_output_deref_op(self) -> Op | None:
        """Get the cached Op for output dereference operation, or None if not supported."""
        if self._output_deref_op is None:
            self._output_deref_op = self._make_output_deref_op()
        return self._output_deref_op

    @property
    def is_input_iterator(self) -> bool:
        """Return True if this iterator supports input dereference."""
        return self.get_input_deref_op() is not None

    @property
    def is_output_iterator(self) -> bool:
        """Return True if this iterator supports output dereference."""
        return self.get_output_deref_op() is not None

    def to_cccl_iter(self, is_output: bool = False) -> Iterator:
        """
        Convert this iterator to a CCCL Iterator for algorithm interop.

        Args:
            is_output: If True, use output_dereference; otherwise use input_dereference

        Returns:
            CCCL Iterator object
        """
        # Get advance op
        advance_op = self.get_advance_op()

        # Get dereference op based on direction
        if is_output:
            deref_op = self.get_output_deref_op()
            if deref_op is None:
                raise ValueError("This iterator does not support output operations")
        else:
            deref_op = self.get_input_deref_op()
            if deref_op is None:
                raise ValueError("This iterator does not support input operations")

        # Create the CCCL Iterator
        return Iterator(
            self._state_alignment,
            IteratorKind.ITERATOR,
            advance_op,
            deref_op,
            self._value_type.info,
            state=self.state,
        )

    @property
    def kind(self) -> Hashable:
        """Return a hashable kind for caching purposes.

        Note: state_bytes is intentionally excluded - iterators with the same
        type structure but different runtime state should share cached reducers.
        """
        return (type(self).__name__, self._value_type)

    # Abstract methods for subclasses
    def _make_advance_op(self) -> Op:
        """
        Create Op object for advance operation.

        Returns:
            Op object with compiled LTOIR
        """
        raise NotImplementedError

    def _make_input_deref_op(self) -> Op | None:
        """
        Create Op object for input dereference operation.

        Returns:
            Op object with compiled LTOIR, or None if not supported
        """
        raise NotImplementedError

    def _make_output_deref_op(self) -> Op | None:
        """
        Create Op object for output dereference operation.

        Returns:
            Op object with compiled LTOIR, or None if not supported
        """
        raise NotImplementedError


def _deterministic_suffix(kind: Hashable) -> str:
    kind_str = str(kind)
    return hashlib.sha256(kind_str.encode()).hexdigest()[:16]


def compose_iterator_states(
    iterators: list[IteratorBase],
) -> tuple[bytes, int, list[int]]:
    """
    Concatenate multiple iterator states with proper alignment.

    This is used by composite iterators (like ZipIterator and PermutationIterator)
    that need to store multiple child iterator states in their own state.

    Args:
        iterators: List of child iterators whose states should be composed

    Returns:
        Tuple of:
        - combined_state_bytes: Concatenated state bytes with padding
        - combined_alignment: Maximum alignment requirement
        - offsets: List of byte offsets for each iterator's state
    """
    if not iterators:
        return (b"", 1, [])

    states = [bytes(memoryview(it.state)) for it in iterators]
    alignments = [it.state_alignment for it in iterators]

    offsets = []
    current_offset = 0
    combined = b""

    for state, align in zip(states, alignments):
        # Add padding to meet alignment requirement
        padding = (align - (current_offset % align)) % align
        combined += b"\x00" * padding
        current_offset += padding

        offsets.append(current_offset)
        combined += state
        current_offset += len(state)

    max_alignment = max(alignments)
    return (combined, max_alignment, offsets)


cache_with_registered_key_functions.register(IteratorBase, lambda it: it.kind)

__all__ = ["IteratorBase"]



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import numba

from .._common import make_binary_tempfile
from .._types import (
    Algorithm,
    Constant,
    Dependency,
    DependentArray,
    DependentPythonOperator,
    Invocable,
    Pointer,
    TemplateParameter,
    numba_type_to_wrapper,
)


def make_merge_sort_keys(
    dtype, items_per_thread, compare_op, threads_in_warp=32, methods=None
):
    """
    Creates a warp-wide merge sort primitive over a
    :ref:`blocked arrangement <flexible-data-arrangement>` of keys.

    Example:
        The code snippet below illustrates a sort of 128 integer keys that
        are partitioned in a
        :ref:`blocked arrangement <flexible-data-arrangement>` across a
        warp of 32 threads where each thread owns 4 consecutive keys.

        The following snippet shows how to invoke the returned
        ``warp_merge_sort`` primitive:

        .. literalinclude:: ../../python/cuda_cccl/tests/coop/test_warp_merge_sort_api.py
            :language: python
            :dedent:
            :start-after: example-begin merge-sort
            :end-before: example-end merge-sort

        Suppose the set of input ``thread_keys`` across the warp of
        threads is ``{ [0, 1, 2, 3], [4, 5, 6, 7], ..., [124, 125, 126, 127] }``.
        The corresponding output ``thread_keys`` in those threads will be
        ``{ [127, 126, 125, 124], [123, 122, 121, 120], ..., [3, 2, 1, 0] }``.

    Args:
        dtype: Numba data type of the keys to be sorted
        threads_in_warp: The number of threads in a warp
        items_per_thread: The number of items each thread owns
        compare_op: Comparison function object. Returns true if the first
            argument is ordered before the second.

    Returns:
        A callable object that can be linked to and invoked from a CUDA
        kernel
    """
    template = Algorithm(
        "WarpMergeSort",
        "Sort",
        "warp_merge_sort",
        ["cub/warp/warp_merge_sort.cuh"],
        [
            TemplateParameter("KeyT"),
            TemplateParameter("ITEMS_PER_THREAD"),
            TemplateParameter("VIRTUAL_WARP_THREADS"),
        ],
        [
            [
                Pointer(numba.uint8),
                DependentArray(Dependency("KeyT"), Dependency("ITEMS_PER_THREAD")),
                DependentPythonOperator(
                    Constant(numba.int8),
                    [Dependency("KeyT"), Dependency("KeyT")],
                    Dependency("Op"),
                ),
            ]
        ],
        type_definitions=[numba_type_to_wrapper(dtype, methods=methods)],
    )
    specialization = template.specialize(
        {
            "KeyT": dtype,
            "VIRTUAL_WARP_THREADS": threads_in_warp,
            "ITEMS_PER_THREAD": items_per_thread,
            "Op": compare_op,
        }
    )
    return Invocable(
        temp_files=[
            make_binary_tempfile(ltoir, ".ltoir")
            for ltoir in specialization.get_lto_ir(threads=threads_in_warp)
        ],
        temp_storage_bytes=specialization.temp_storage_bytes,
        temp_storage_alignment=specialization.temp_storage_alignment,
        algorithm=specialization,
    )



# Copyright (c) 2024, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# SPDX-License  -Identifier: Apache-2.0 WITH LLVM-exception

from . import experimental

__all__ = [
    "experimental",
]



# Copyright (c) 2024-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from typing import Union

import numpy as np

from .. import _bindings
from .. import _cccl_interop as cccl
from .._caching import cache_with_registered_key_functions
from .._cccl_interop import call_build, set_cccl_iterator_state, to_cccl_value_state
from .._utils.protocols import get_data_pointer, validate_and_get_stream
from .._utils.temp_storage_buffer import TempStorageBuffer
from ..typing import DeviceArrayLike, IteratorT


class _Histogram:
    __slots__ = [
        "num_rows",
        "d_samples_cccl",
        "d_histogram_cccl",
        "h_num_output_levels_cccl",
        "h_lower_level_cccl",
        "h_upper_level_cccl",
        "build_result",
    ]

    def __init__(
        self,
        d_samples: DeviceArrayLike | IteratorT,
        d_histogram: DeviceArrayLike,
        h_num_output_levels: np.ndarray,
        h_lower_level: np.ndarray,
        h_upper_level: np.ndarray,
        num_samples: int,
    ):
        num_channels = 1
        num_active_channels = 1
        is_evenly_segmented = True
        self.num_rows = 1
        num_levels = h_num_output_levels[0]
        row_stride_samples = num_samples

        self.d_samples_cccl = cccl.to_cccl_input_iter(d_samples)
        self.d_histogram_cccl = cccl.to_cccl_output_iter(d_histogram)
        self.h_num_output_levels_cccl = cccl.to_cccl_value(h_num_output_levels)
        self.h_lower_level_cccl = cccl.to_cccl_value(h_lower_level)
        self.h_upper_level_cccl = cccl.to_cccl_value(h_upper_level)

        self.build_result = call_build(
            _bindings.DeviceHistogramBuildResult,
            num_channels,
            num_active_channels,
            self.d_samples_cccl,
            num_levels,
            self.d_histogram_cccl,
            self.h_lower_level_cccl,
            self.num_rows,
            row_stride_samples,
            is_evenly_segmented,
        )

    def __call__(
        self,
        temp_storage,
        d_samples: DeviceArrayLike | IteratorT,
        d_histogram: DeviceArrayLike,
        h_num_output_levels: np.ndarray,
        h_lower_level: np.ndarray,
        h_upper_level: np.ndarray,
        num_samples: int,
        stream=None,
    ):
        set_cccl_iterator_state(self.d_samples_cccl, d_samples)
        set_cccl_iterator_state(self.d_histogram_cccl, d_histogram)
        self.h_num_output_levels_cccl.state = to_cccl_value_state(h_num_output_levels)
        self.h_lower_level_cccl.state = to_cccl_value_state(h_lower_level)
        self.h_upper_level_cccl.state = to_cccl_value_state(h_upper_level)

        stream_handle = validate_and_get_stream(stream)
        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            # Note: this is slightly slower, but supports all ndarray-like objects as long as they support CAI
            # TODO: switch to use gpumemoryview once it's ready
            d_temp_storage = get_data_pointer(temp_storage)

        temp_storage_bytes = self.build_result.compute_even(
            d_temp_storage,
            temp_storage_bytes,
            self.d_samples_cccl,
            self.d_histogram_cccl,
            self.h_num_output_levels_cccl,
            self.h_lower_level_cccl,
            self.h_upper_level_cccl,
            num_samples,
            self.num_rows,
            num_samples,
            stream_handle,
        )

        return temp_storage_bytes


@cache_with_registered_key_functions
def _make_histogram_even_impl(
    d_samples: DeviceArrayLike | IteratorT,
    d_histogram: DeviceArrayLike,
    num_output_levels_val: int,
    lower_level_val,
    upper_level_val,
    level_dtype,
    num_samples: int,
    uses_privatized_smem: bool,
):
    """Internal cached implementation of make_histogram_even.

    The uses_privatized_smem parameter ensures kernels compiled
    for different bin count regimes aren't reused.
    """
    # Reconstruct the numpy arrays expected by _Histogram
    h_num_output_levels = np.array([num_output_levels_val], dtype=np.int32)
    h_lower_level = np.array([lower_level_val], dtype=level_dtype)
    h_upper_level = np.array([upper_level_val], dtype=level_dtype)

    return _Histogram(
        d_samples,
        d_histogram,
        h_num_output_levels,
        h_lower_level,
        h_upper_level,
        num_samples,
    )


def make_histogram_even(
    d_samples: DeviceArrayLike | IteratorT,
    d_histogram: DeviceArrayLike,
    h_num_output_levels: np.ndarray,
    h_lower_level: np.ndarray,
    h_upper_level: np.ndarray,
    num_samples: int,
):
    """Implements a device-wide histogram that places ``d_samples`` into evenly-spaced bins.

    Example:
        Below, ``make_histogram_even`` is used to create a histogram object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/histogram/histogram_object.py
          :language: python
          :start-after: # example-begin

    Args:
        d_samples: Device array or iterator containing the input samples to be histogrammed
        d_histogram: Device array to store the histogram
        h_num_output_levels: Host array containing the number of output levels
        h_lower_level: Host array containing the lower level
        h_upper_level: Host array containing the upper level
        num_samples: Number of samples to be histogrammed

    Returns:
        A callable object that can be used to perform the histogram
    """
    # Extract scalar values from arrays for caching
    num_output_levels_val = int(h_num_output_levels[0])
    lower_level_val = h_lower_level[0].item()
    upper_level_val = h_upper_level[0].item()
    level_dtype = h_lower_level.dtype

    # bins <= 256 uses privatized smem strategy. a different compile
    # path than bins > 256. We should include this information when
    # caching histogram build objects.
    # See detail::histogram::max_privatized_smem_bins (dispatch_histogram.cuh)
    num_bins = num_output_levels_val - 1
    uses_privatized_smem = num_bins <= 256
    return _make_histogram_even_impl(
        d_samples,
        d_histogram,
        num_output_levels_val,
        lower_level_val,
        upper_level_val,
        level_dtype,
        num_samples,
        uses_privatized_smem,
    )


def histogram_even(
    d_samples: DeviceArrayLike | IteratorT,
    d_histogram: DeviceArrayLike,
    num_output_levels: int,
    lower_level: Union[np.floating, np.integer],
    upper_level: Union[np.floating, np.integer],
    num_samples: int,
    stream=None,
):
    """
    Performs device-wide histogram computation with evenly-spaced bins.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``histogram_even`` is used to compute a histogram with evenly-spaced bins.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/histogram/histogram_even_basic.py
            :language: python
            :start-after: # example-begin
            :caption: Basic histogram example.

    Args:
        d_samples: Device array or iterator containing the input sequence of data samples
        d_histogram: Device array to store the computed histogram
        num_output_levels: Number of histogram bin levels (num_bins = num_output_levels - 1)
        lower_level: Lower sample value bound (inclusive)
        upper_level: Upper sample value bound (exclusive)
        num_samples: Number of input samples
        stream: CUDA stream for the operation (optional)
    """
    # Histogram can accept multiple channels, with one value per channel for
    # each of these parameters. The API only supports one channel for now but we
    # pass arrays to make_histogram_even to support multiple channels in the
    # future.
    h_num_output_levels = np.array([num_output_levels], dtype=np.int32)
    h_lower_level = np.array([lower_level], dtype=type(lower_level))
    h_upper_level = np.array([upper_level], dtype=type(upper_level))

    histogram = make_histogram_even(
        d_samples,
        d_histogram,
        h_num_output_levels,
        h_lower_level,
        h_upper_level,
        num_samples,
    )
    temp_storage_bytes = histogram(
        None,
        d_samples,
        d_histogram,
        h_num_output_levels,
        h_lower_level,
        h_upper_level,
        num_samples,
        stream,
    )
    temp_storage = TempStorageBuffer(temp_storage_bytes, stream)
    histogram(
        temp_storage,
        d_samples,
        d_histogram,
        h_num_output_levels,
        h_lower_level,
        h_upper_level,
        num_samples,
        stream,
    )



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
#
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from typing import Callable

import numpy as np

from .. import _bindings
from .. import _cccl_interop as cccl
from .._caching import cache_with_registered_key_functions
from .._cccl_interop import (
    call_build,
    get_value_type,
    set_cccl_iterator_state,
    to_cccl_value_state,
)
from .._utils.protocols import (
    get_data_pointer,
    validate_and_get_stream,
)
from .._utils.temp_storage_buffer import TempStorageBuffer
from ..op import OpAdapter, make_op_adapter
from ..typing import DeviceArrayLike, GpuStruct, IteratorT, Operator


class _SegmentedReduce:
    __slots__ = [
        "build_result",
        "d_in_cccl",
        "d_out_cccl",
        "start_offsets_in_cccl",
        "end_offsets_in_cccl",
        "h_init_cccl",
        "op_cccl",
    ]

    def __init__(
        self,
        d_in: DeviceArrayLike | IteratorT,
        d_out: DeviceArrayLike | IteratorT,
        start_offsets_in: DeviceArrayLike | IteratorT,
        end_offsets_in: DeviceArrayLike | IteratorT,
        op: OpAdapter,
        h_init: np.ndarray | GpuStruct,
    ):
        self.d_in_cccl = cccl.to_cccl_input_iter(d_in)
        self.d_out_cccl = cccl.to_cccl_output_iter(d_out)
        self.start_offsets_in_cccl = cccl.to_cccl_input_iter(start_offsets_in)
        self.end_offsets_in_cccl = cccl.to_cccl_input_iter(end_offsets_in)
        self.h_init_cccl = cccl.to_cccl_value(h_init)

        # Compile the op with value types
        value_type = get_value_type(h_init)

        self.op_cccl = op.compile((value_type, value_type), value_type)

        self.build_result = call_build(
            _bindings.DeviceSegmentedReduceBuildResult,
            self.d_in_cccl,
            self.d_out_cccl,
            self.start_offsets_in_cccl,
            self.end_offsets_in_cccl,
            self.op_cccl,
            self.h_init_cccl,
        )

    def __call__(
        self,
        temp_storage,
        d_in,
        d_out,
        op: Callable | OpAdapter,
        num_segments: int,
        start_offsets_in,
        end_offsets_in,
        h_init,
        max_segment_size: int | None = None,
        stream=None,
    ):
        if num_segments > np.iinfo(np.int32).max:
            raise RuntimeError(
                "Segmented sort does not currently support more than 2^31-1 segments."
            )

        if max_segment_size is None:
            max_segment_size = 0  # CCCL.c treats 0 as "not specified"

        set_cccl_iterator_state(self.d_in_cccl, d_in)
        set_cccl_iterator_state(self.d_out_cccl, d_out)
        set_cccl_iterator_state(self.start_offsets_in_cccl, start_offsets_in)
        set_cccl_iterator_state(self.end_offsets_in_cccl, end_offsets_in)

        op_adapter = make_op_adapter(op)
        self.op_cccl.state = op_adapter.get_state()

        self.h_init_cccl.state = to_cccl_value_state(h_init)

        stream_handle = validate_and_get_stream(stream)

        if temp_storage is None:
            temp_storage_bytes = 0
            d_temp_storage = 0
        else:
            temp_storage_bytes = temp_storage.nbytes
            d_temp_storage = get_data_pointer(temp_storage)

        temp_storage_bytes = self.build_result.compute(
            d_temp_storage,
            temp_storage_bytes,
            self.d_in_cccl,
            self.d_out_cccl,
            num_segments,
            self.start_offsets_in_cccl,
            self.end_offsets_in_cccl,
            self.op_cccl,
            self.h_init_cccl,
            max_segment_size,
            stream_handle,
        )
        return temp_storage_bytes


@cache_with_registered_key_functions
def make_segmented_reduce(
    d_in: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    start_offsets_in: DeviceArrayLike | IteratorT,
    end_offsets_in: DeviceArrayLike | IteratorT,
    op: Operator,
    h_init: np.ndarray | GpuStruct,
):
    """Computes a device-wide segmented reduction using the specified binary ``op`` and initial value ``init``.

    Example:
        Below, ``make_segmented_reduce`` is used to create a segmented reduction object that can be reused.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/segmented/segmented_reduce_object.py
            :language: python
            :start-after: # example-begin


    Args:
        d_in: Device array or iterator containing the input sequence of data items
        d_out: Device array that will store the result of the reduction
        start_offsets_in: Device array or iterator containing offsets to start of segments
        end_offsets_in: Device array or iterator containing offsets to end of segments
        op: Binary operator to apply.
            The signature is ``(T, T) -> T``, where ``T`` is
            the data type of the initial value ``h_init``.
        init: Numpy array storing initial value of the reduction

    Returns:
        A callable object that can be used to perform the reduction
    """
    op_adapter = make_op_adapter(op)
    return _SegmentedReduce(
        d_in, d_out, start_offsets_in, end_offsets_in, op_adapter, h_init
    )


def segmented_reduce(
    d_in: DeviceArrayLike | IteratorT,
    d_out: DeviceArrayLike | IteratorT,
    start_offsets_in: DeviceArrayLike | IteratorT,
    end_offsets_in: DeviceArrayLike | IteratorT,
    op: Operator,
    h_init: np.ndarray | GpuStruct,
    num_segments: int,
    max_segment_size: int | None = None,
    stream=None,
):
    """
    Performs device-wide segmented reduction.

    This function automatically handles temporary storage allocation and execution.

    Example:
        Below, ``segmented_reduce`` is used to compute the minimum value of segments in a sequence of integers.

        .. literalinclude:: ../../python/cuda_cccl/tests/compute/examples/segmented/segmented_reduce_basic.py
            :language: python
            :start-after: # example-begin


    Args:
        d_in: Device array or iterator containing the input sequence of data items
        d_out: Device array to store the result of the reduction for each segment
        start_offsets_in: Device array or iterator containing the sequence of beginning offsets
        end_offsets_in: Device array or iterator containing the sequence of ending offsets
        op: Binary operator to apply.
            The signature is ``(T, T) -> T``, where ``T`` is
            the data type of the initial value ``h_init``.
        h_init: Initial value for the reduction
        num_segments: Number of segments to reduce
        max_segment_size: The number of elements in the largest segment (optional)
            If provided, this information is used to dispatch to the
            optimal kernel for best performance.
        stream: CUDA stream for the operation (optional)
    """
    reducer = make_segmented_reduce(
        d_in, d_out, start_offsets_in, end_offsets_in, op, h_init
    )
    tmp_storage_bytes = reducer(
        None,
        d_in,
        d_out,
        op,
        num_segments,
        start_offsets_in,
        end_offsets_in,
        h_init,
        max_segment_size,
        stream,
    )
    tmp_storage = TempStorageBuffer(tmp_storage_bytes, stream)
    reducer(
        tmp_storage,
        d_in,
        d_out,
        op,
        num_segments,
        start_offsets_in,
        end_offsets_in,
        h_init,
        max_segment_size,
        stream,
    )
